/******************************************************************************
File:    cpc.c
Purpose: Sample C language control program for Comtrol Hostess i/S2.
         Supports asynchronous, HDLC, and Bisync communications.
Company: Comtrol Corporation
Author:  Craig Harrison
Ident:   $Id$
*******************************************************************************
Copyright 1994 Comtrol Corporation. All rights reserved.
Subject to developers license agreement.
*******************************************************************************
                              WARNINGS

1. All ISRs should call ISR_PROLOG as their first instruction, and call
ISR_EPILOG as their last instruction.  Some ISRs may use an assembly
language version of ISR_EPILOG to improve performance.

2. Do not disable or enable interrupts directly.  Instead use macros DIS_INT
and EN_INT.  These macros, together with ISR_PROLOG and ISR_EPILOG, will
prevent any function from enabling interrupts prematurely if called from
within an ISR.

3. All accesses to SCC registers that are paired because of indexing must be
protected by disabling interrupts.  If interrupts are already disabled (as
when in an interrupt service routine) they should not be disabled/enabled
again.  When the ISR exits, the iret instruction will automatically reenable
interrupts.

4. If modifying this program, do not make calls to any C library functions that
require DOS support.  DOS is not running on the controller.

5. If using the controller's firmware debugger, interrupt service routines must
not access any near memory.  All pointers must be far pointers, and all
variables must be far.  This is required because the debugger modifies the
DS register while operating, but ISRs will continue to interrupt the debugger.
Therefore ISRs cannot assume that the DS is unchanged while the debugger is
running.

6. When disabling/enabling a DMA channel, the sequence:
      Read DMK (DMA mask register)
      Modify value read
      Write value to DMK
is used.  This sequence must be protected by disabling interrupts
and disabling the DMA controller, as shown in macros DIS_DMA and EN_DMA.
This is because when the DMA controller detects a terminal count it
modifies the DMK.  If this occurs after the DMK read but before the
DMK write the value written could be invalid.

7. When accessing a double word that can be modified by an ISR disable the
interrupts (see GetCtlTime()). This is because double word transfers are
not atomic, so the ISR could modify the value between the word two word
transfers required to access the double word.

8. When accessing a value from dual port memory using a non-atomic operation,
if that value can be modified by the system that operation must be protected.
Some non-atomic operations are:
   1. Double word reads
   2. Any read - modify - write
A way to protect any non-atomic access is to use a semaphore (see GET_DPSEM
and REL_DPSEM). A way to protect a non-atomic read is to repeatadly read the
value until subsequent reads match.

******************************************************************************/


#include <dos.h>

#ifdef BORLANDC
#include <mem.h>
#endif

#ifdef MSC
#include <memory.h>
#endif

#include <string.h>
#include "cpc.h"
#include "dpram.h"

extern UWORD_T endcp_seg;              // segment immediately after ctrl pgm

BYTE_T far *pHeap;                     // ptr to heap.  This is area following
                                       // control program, it is available for
                                       // program useage.
FIRMUSER_T far *pFirm;                 // ptr to firmware user area
MSGQ_T far *pCtlQ;                     // ptr to CTL message queue
MSGQ_T far *pSysQ;                     // ptr to SYS message queue
WORD_T far *pTrBuf;                    // ptr to trace buffer
WORD_T far *pTrIx;                     // ptr to trace buffer index
BYTE_T far *pDpSemCtlAccess;           // ptr to dual port semaphore
BYTE_T far *pDpSemSysAccess;           // ptr to dual port semaphore
UWORD_T far *pSfg;                     // ptr to Status Flag Groups
WORD_T SysIntPend;                     // set if an int to system is pending
LINE_T far *Line[NUMLINES];            // ptrs to line structures
WORD_T InDispatcher = 0;               // dispatcher executing flag
WORD_T InIsr = 0;                      // ISR executing flag
UWORD_T IntDis = 0;                    // >0 indicates ints disabled
BYTE_T SysMsg[MSG_LEN];                // SYS msg buffer, use only at int time

// Table of SCC baud rate values.  The MSG_BAUD_X flags are used as indices
// into this table.
WORD_T BaudTbl[MSG_BAUD_115200+1] =
{
   BPS50,BPS75,BPS110,BPS134,BPS150,BPS300,BPS600,BPS1200,BPS1800,BPS2000,
   BPS2400,BPS3600,BPS4800,BPS7200,BPS9600,BPS19200,BPS38400,BPS57600,
   BPS76800,BPS115200
};

// Table of SCC parity values.  The MSG_PAR_X flags are used as indices
// into this table.
BYTE_T ParityTbl[MSG_PAR_EVEN+1] =
{
   0,PARITY_ENABLE | PARITY_ODD,PARITY_ENABLE | PARITY_EVEN};

// Table of SCC stop bit values.  The MSG_STOPBIT_X flags are used as indices
// into this table.
BYTE_T StopBitTbl[MSG_STOPBIT_2+1] =
{ONE_STOP_BIT,ONEHALF_STOP_BITS,TWO_STOP_BITS};

// Table of SCC Rx data bit values.  The MSG_DATABIT_X flags are used as indices
// into this table.
BYTE_T RxDataBitTbl[MSG_DATABIT_8+1] =
{
   RX5_BIT_CHAR,RX6_BIT_CHAR,RX7_BIT_CHAR,RX8_BIT_CHAR
};

// Table of SCC Tx data bit values.  The MSG_DATABIT_X flags are used as indices
// into this table.
BYTE_T TxDataBitTbl[MSG_DATABIT_8+1] =
{
   TX5_BIT_CHAR,TX6_BIT_CHAR,TX7_BIT_CHAR,TX8_BIT_CHAR
};


/***************************************************************************
Function: main
Purpose:  Initialization and main dispatching loop.
Comments: This is called from module cpcstart.asm, it never returns.
*/
main()
{
WORD_T nLn;                            // line number
LINE_T far *pLn;                       // ptr to line structure
WORD_T i;
static BYTE_T Msg[MSG_LEN];            // SYS msg buf, for non-int time msgs
#if LN0_ASYNC || LN1_ASYNC
BYTE_T b;
WORD_T In;                             // Tx queue In index
#endif

InitRam();                             // initialize RAM
InitDmaCtrl();                         // initialize DMA controller
InitTimer(1,CNT200,Timer1Isr);         // start up timer 1, 5ms period
ExecTestModes();                       // execute special tests modes
Msg[0] = 0;                            // prevent compiler warning


// Main dispatching loop.  This handles all low priority processing, that is,
// everything that does not happen at interrupt time.
while(1)
{
   if(pCtlQ->In != pCtlQ->Out)         // messages in controller msg queue
      DispatchMsg();                   // dispatch them

   if(SysIntPend)                      // int to system is pending
   {
      SysIntPend = 0;
      OUTB(C2SINT_IO,C2SINT_LOW);      // set interrupt line low
      for(i = 0;i < 50;i++)            // delay to ensure host sees transition
         i = i;
      OUTB(C2SINT_IO,C2SINT_HI);       // set interrupt line high
   }

   for(nLn = 0;nLn < NUMLINES;nLn++)   // processing for each line
   {
      pLn = Line[nLn];                 // get ptr to line structure

#if LN0_ASYNC || LN1_ASYNC             // *** Asynchronous processing ***
      // For asynchronous Tx is polled and is handled here, but Rx is interrupt
      // driven and is handled in LnXRcaIsr().
      if(pLn->Status & CLOSE_PEND)     // close pending
      {
         // Wait for Tx queue to empty before closing so don't cut off
         // output in middle of a message.  If timeout occurs before empty
         // continue with close anyway.
         if(GetCtlTime() > pLn->Timeout) // timed out
            Close(pLn);                // go ahead and close
         else if(pLn->TxqIn == pLn->TxqOut) // Tx queue empty
         {
            // Make sure last bit has been sent before closing to allow the
            // last byte to completely shift out
            DIS_INT();
            OUTB(pLn->BaseIo,RR1);
            b = INB(pLn->BaseIo);
            EN_INT();
            if(b & ALLSENT)            // last bit is gone
               Close(pLn);             // finally - OK to close
         }
      }

      if(pLn->Status & ASYNC_OPEN)     // line is open for async
      {
         // Polled async transmit data processing.  This is more efficient
         // than interrupt driven.
         In = GetTxqIn(pLn);
         i = pLn->TxqOut;
         if(In != i)                   // Tx queue has data available
         {
            if(INB(pLn->BaseIo) & TX_BUFFER) // SCC Tx buffer is empty
            {
               OUTB(pLn->BaseIo + 2,*(pLn->TxqCtl + i)); // write char to SCC
#if !TEST_ASYNC                       // in async test mode never remove char
               pLn->TxqOut = (i + 1) % TXQ_SIZE; // update Out pointer
#endif // end #if !TEST_ASYNC
            }
         }
      }
#endif // end asynchronous processing


#if LN0_HDLC || LN1_HDLC               // *** HDLC processing ***
      // For HDLC both Tx and Rx use DMA. Rx DMA is controlled by a state
      // machine implemented here and in LnXSrcIsr().  Tx DMA is controlled
      // by a state machine implemented here, in DmaTxIsr(), and in
      // LnXEscIsr().
      if(pLn->Status & HDLC_OPEN)      // line is open for HDLC
      {
         if(pLn->RxState == HRX_BUF_WAIT) // Rx state machine
         {
            if(pLn->RxbqCnt < RXBQ_SIZE) // a buffer is available
            {
               StartRxHdlc(pLn);       // start Rx
            }
         }

         switch(pLn->TxState)          // Tx state machine
         {
            case STX_IDLE:             // Tx idle state
               if(pLn->Status & CLOSE_PEND) // close pending
               {
                  // Wait for timeout before closing to allow SYNC char
                  // after frame to transmit
                  if(GetCtlTime() > pLn->Timeout) // timed out
                     Close(pLn);
               }
               else if(pLn->TxbqCnt > 0) // a frame is available to Tx
                  StartTxSync(pLn);    // start Tx DMA
               break;
            case STX_ACTIVE:           // Tx DMA in process. TC int or ESC int
                                       //    with Tx Underrun exits state.
               break;
            case STX_RETRY:            // Tx frame aborted
               if(pLn->Status & CLOSE_PEND) // close pending
                  Close(pLn);
               else
                  RetryTxHdlc(pLn);    // retry sending frame
               break;
            case STX_CLOSE_WAIT:       // wait for SYS to close line.
               if(pLn->Status & CLOSE_PEND) // close pending
                  Close(pLn);
               break;
            case STX_EOF_WAIT:         // wait last info byte in frame to Tx.
                                       //    ESC Tx Underrun exits state.
               break;
            case STX_EOF:              // end of frame processing
               // Wait to do end of frame processing until the CRC is
               // sent, which is indicated by SCC Tx buffer empty.
               // If we don't wait till the CRC is gone and another frame is
               // immediately available the CRC will be corrupted by
               // StartTxSync().
               if(INB(pLn->BaseIo) & TX_BUFFER) // CRC completely sent
               {
                  // Update Tx buffer queue Out index and count.  Need to
                  // protect TxbCnt-- with dual port semaphore because the --
                  // operation is not atomic and there is the possibliity that
                  // both system and controller could be changing it
                  // simultaneously.
                  DIS_INT();
                  pLn->TxbqOut = (pLn->TxbqOut + 1) % TXBQ_SIZE; // deq the buf
                  GET_DPSEM();
#if !TEST_HDLC
                  pLn->TxbqCnt--;
                  SysIntPend = 1;      // queue an interrupt to the system
#endif
                  REL_DPSEM();
                  pLn->TxState = STX_IDLE; // state machine to Tx idle state
                  EN_INT();
               }
               break;
            default:
               Close(pLn);
               Msg[0] = SYSCMD_STATEERROR; // msg to SYS, command field
               Msg[1] = 0;                 //    line number field
               EnqSysMsg(Msg);             // send message to system
               OUTB(C2SINT_IO,C2SINT_LOW); // set interrupt line low
               for(i = 0;i < 50;i++)       // delay so SYS sees transition
                  i = i;
               OUTB(C2SINT_IO,C2SINT_HI);  // set interrupt line high
         }
      }
#endif // end HDLC processing


#if LN0_BISYNC || LN1_BISYNC           // *** BISYNC processing ***
      // For Bisync Tx uses DMA and is controlled by a state machine
      // implemented here, in DmaTxIsr(), and in LnXEscIsr().  Rx is interrupt
      // driven and is controlled by a state machine implemented here, in
      // LnXSrcIsr(), and in LnXRcaIsr().
      if(pLn->Status & BISYNC_OPEN)    // line is open for Bisync
      {
         if(pLn->RxState == BRX_BUF_WAIT) // Rx state machine
         {
            if(pLn->RxbqCnt < RXBQ_SIZE) // a buffer is available
            {
               StartRxBisync(pLn);     // start Rx
            }
         }

         switch(pLn->TxState)          // Tx state machine
         {
            case STX_IDLE:             // Tx idle state
               if(pLn->Status & CLOSE_PEND) // close pending
               {
                  // Wait for timeout before closing to allow SYNC char
                  // after frame to transmit
                  if(GetCtlTime() > pLn->Timeout) // timed out
                     Close(pLn);
               }
               else if(pLn->TxbqCnt > 0) // a frame is available to Tx
               {
                     StartTxSync(pLn);    // start Tx DMA
               }
               break;
            case STX_ACTIVE:           // Tx DMA in process. TC int or ESC int
                                       //    with Tx Underrun exits state.
               break;
            case STX_RETRY:            // Tx frame aborted
               if(pLn->Status & CLOSE_PEND) // close pending
                  Close(pLn);
               else
                  RetryTxHdlc(pLn);    // retry sending frame
               break;
            case STX_CLOSE_WAIT:       // wait for SYS to close line.
               if(pLn->Status & CLOSE_PEND) // close pending
                  Close(pLn);
               break;
            case STX_EOF_WAIT:         // wait last info byte in frame to Tx.
                                       //    ESC Tx Underrun exits state.
               break;
            case STX_EOF:              // end of frame processing
               // Wait to do till end of frame processing until the CRC is
               // sent, which is indicated by SCC Tx buffer empty.
               // If we don't wait till the CRC is gone and another frame is
               // immediately available the CRC will be corrupted by
               // StartTxSync().
               if(INB(pLn->BaseIo) & TX_BUFFER) // CRC completely sent
               {
                  // Update Tx buffer queue Out index and count.  Need to
                  // protect TxbCnt-- with dual port semaphore because the --
                  // operation is not atomic and there is the possibliity that
                  // both system and controller could be changing it
                  // simultaneously.
                  DIS_INT();
                  pLn->TxbqOut = (pLn->TxbqOut + 1) % TXBQ_SIZE; // deq the buf
                  GET_DPSEM();
                  pLn->TxbqCnt--;
                  REL_DPSEM();
                  pLn->TxState = STX_IDLE; // state machine to Tx idle state
                  EN_INT();
                  SysIntPend = 1;      // queue an interrupt to the system
               }
               break;
            default:
               Close(pLn);
               Msg[0] = SYSCMD_STATEERROR; // msg to SYS, command field
               Msg[1] = 0;                 //    line number field
               EnqSysMsg(Msg);             // send message to system
               OUTB(C2SINT_IO,C2SINT_LOW); // set interrupt line low
               for(i = 0;i < 50;i++)       // delay so SYS sees transition
                  i = i;
               OUTB(C2SINT_IO,C2SINT_HI);  // set interrupt line high
         }
      }
#endif // end BISYNC processing

   } // end for(nLn...)
} // end while(1)
} // end main()

/***************************************************************************
Function: InitRam
Purpose:  Initialize RAM. This includes pointers to queues and line structures,
          interrupt vectors, line structures, and the firmware data area
          interaction flag.
Call:     InitRam()
Return:   void
*/
void InitRam(void)
{
   WORD_T nLn;                         // line number
   WORD_T nBuf;                        // buffer number
   LINE_T far *pLn;                    // ptr to line structure
   DPM_T far *pDpm;                    // ptr to DPM_T structure
   WORD_T VectorType;                  // interrupt vector type
   UDWORD_T AbsAddr;                   // an absolute address
   static WORD_T IoAddr[] =            // SCC I/O addresses
   {
      LN00_IO, LN01_IO
   };

   // Initialize trace buffer pointers
   DIS_INT();                    // disable interrupts while initializing RAM
   pDpm = (DPM_T far *)0;             // DPM_T struct start a bottom of memory
   pTrBuf = (WORD_T far *)pDpm->TrBuf;
   pTrIx = &pDpm->TrIx;

   // Initialize interrupt vector table
   InitVector();

   // Clear DPM_T struct except for vector and firmware data areas
   _fmemset(&pDpm->TrBuf,0,sizeof(DPM_T) - (VECTAREA_SIZE+FIRMDATA_SIZE));

   // Initialize message queues
   pCtlQ = &pDpm->CtlQ;
   pSysQ = &pDpm->SysQ;

   // Initialize ptrs to dual port semaphore
   pDpSemCtlAccess = &pDpm->DpSem.CtlAccess;
   pDpSemSysAccess = &pDpm->DpSem.SysAccess;

   // Initialize ptr to Status Flag Groups
   pSfg = pDpm->Sfg;

   // Initialize line structures to non-zero defaults
   VectorType = SCCBASE_TYPE;
   for(nLn = 0;nLn < NUMLINES;nLn++)
   {
      pLn = &pDpm->Line[nLn];
      Line[nLn] = pLn;
      pLn->LineNum = nLn;
      pLn->BaseIo = IoAddr[nLn];
      pLn->WR2_ = VectorType;     // both SCC channels get same base vector
      VectorType += (nLn % 2) * 0x10;
      pLn->TxqCtl = pDpm->TxqBuf[nLn];
      pLn->RxqCtl = pDpm->RxqBuf[nLn];
      pLn->TxqSys = pDpm->TxqBuf[nLn];  // system must adjust this
      pLn->RxqSys = pDpm->RxqBuf[nLn];  // system must adjust this
      for(nBuf = 0;nBuf < TXBQ_SIZE;nBuf++)
      {
         pLn->Txb[nBuf].BufCtl = pDpm->TxbqBuf[nLn][nBuf];
         AbsAddr = FarPtrToAbs(pLn->Txb[nBuf].BufCtl);  //absolute addr
         pLn->Txb[nBuf].BufLo = (UWORD_T)AbsAddr; // low word of absolute addr
         pLn->Txb[nBuf].BufHi = (UWORD_T)(AbsAddr >> 16); // high word
         pLn->Txb[nBuf].BufSys = pDpm->TxbqBuf[nLn][nBuf]; // system must adjust
      }
      for(nBuf = 0;nBuf < RXBQ_SIZE;nBuf++)
      {
         pLn->Rxb[nBuf].BufCtl = pDpm->RxbqBuf[nLn][nBuf];
         AbsAddr = FarPtrToAbs(pLn->Rxb[nBuf].BufCtl);  //absolute addr
         pLn->Rxb[nBuf].BufLo = (UWORD_T)AbsAddr; // low word of absolute addr
         pLn->Rxb[nBuf].BufHi = (UWORD_T)(AbsAddr >> 16); // high word
         pLn->Rxb[nBuf].BufSys = pDpm->RxbqBuf[nLn][nBuf]; // system must adjust
      }
   }
   Line[0]->TxDmaChan = LN0_TXDMACH;   // DMA channel used for Tx
   Line[0]->RxDmaChan = LN0_RXDMACH;   // DMA channel used for Rx
   Line[0]->DisTxDma = LN0_DISTXDMA;   // Tx DMA chan mask for disable
   Line[0]->DisRxDma = LN0_DISRXDMA;   // Tx DMA chan mask for disable
   Line[0]->EnTxDma = LN0_ENTXDMA;     // Tx DMA chan mask for enable
   Line[0]->EnRxDma = LN0_ENRXDMA;     // Tx DMA chan mask for enable
   Line[1]->TxDmaChan = LN1_TXDMACH;   // DMA channel used for Tx */
   Line[1]->RxDmaChan = LN1_RXDMACH;   // DMA channel used for Rx */
   Line[1]->DisTxDma = LN1_DISTXDMA;   // Tx DMA chan mask for disable
   Line[1]->DisRxDma = LN1_DISRXDMA;   // Tx DMA chan mask for disable
   Line[1]->EnTxDma = LN1_ENTXDMA;     // Tx DMA chan mask for enable
   Line[1]->EnRxDma = LN1_ENRXDMA;     // Tx DMA chan mask for enable

   // Initialize configuration control register.
   // This is set up to use alternate DTR so that the SCC DTR/REQ pin can
   // be used as the Tx DMA request line.
   ASM int CFGCTL_READINT
   ASM or   ax,LN0_ALTDTR_EN + LN1_ALTDTR_EN + LN0_ALTDTR + LN1_ALTDTR;
   ASM int CFGCTL_WRITEINT

   // Initialize Tx clock mask values.  These are used to set the Tx clock
   // to either DCE or DTE source through the configuration control register.
   Line[0]->TxClkMask = LN0_TXCLK_DTE;
   Line[1]->TxClkMask = LN1_TXCLK_DTE;

   // Initialize alternate DTR mask values.  These are used to set/clear
   // DTR through the configuration control register.
   Line[0]->AltDtrMask = LN0_ALTDTR;
   Line[1]->AltDtrMask = LN1_ALTDTR;

   // Initialize firmware user area and heap pointers
   pHeap = (BYTE_T far *)MAKE_FP(endcp_seg,0); // heap ptr
   pFirm = &pDpm->FirmUser;            // firm user area ptr

   OUTB(DICM,1);                       // reset DMA controller
   pFirm->i_flag = 0x55aa;             // signal system we're alive
   EN_INT();                           // reenable interrupts
}

/***************************************************************************
Function: InitTimer
Purpose:  Initialize a timer and start it up.
Call:     InitTimer(TNum,Count,Isr)
          WORD_T TNum: Timer number (0, 1, or 2).
          UWORD_T Count: Count value to be written to count register.
          void (interrupt far *Isr)(): ptr to interrupt service routine.
Return:   void
Comments: This function initializes the timer interrupt vector and timer
          registers.
*/
void InitTimer(WORD_T TNum,UWORD_T Count,void (interrupt far *Isr)())
{
   static TIMER_T Timers[MAX_TIMER-MIN_TIMER+1] = // information on each timer
   {
      {TIM0_TYPE, TIM0_CTLWD, TIMER0},
      {TIM1_TYPE, TIM1_CTLWD, TIMER1},
      {TIM2_TYPE, TIM2_CTLWD, TIMER2}
   };

   if(TNum < MIN_TIMER || TNum > MAX_TIMER)  // invalid timer number
      return;

   DIS_INT();          // disable interrupts before modifying vector table
   SetVector(Timers[TNum].VectorType,Isr); // initialize interrupt vector
   OUTB(TIM_CTL,Timers[TNum].ControlWord); // control reg
   OUTB(Timers[TNum].CountReg,Count);      // count least significant byte
   OUTB(Timers[TNum].CountReg,Count >> 8); // count most significant byte
   EN_INT();                           // reenable interrupts
}

/***************************************************************************
Function: InitVector
Purpose:  Initialize interrupt vector table.
Call:     InitVector()
Return:   void
Comments: Each vector requires 4 bytes, two for the offset and two for the
          segment.  The firmware has already initialized all unused vectors
          to point to an "invalid interrupt ISR," so it is not necessary to
          do so again here.  The SCC vectors are contained in the table below.
          Since the SCC modifies bits 3, 2, and 1 of the base vector type, but
          does not modify bit 0, every second SCC vector is unused.

          Timer interrupt vectors are initialized in InitTimer() rather
          than here.

          HDLC does not use a seperate RCA ISR.  Instead, the SRC ISR address
          is placed in the RCA vector.  This is because the "Recevie Interrupt
          on Special Condition only" mode can actually generate an RCA
          interrupt under certian conditions, see Am85C30 SCC 1992 Technical
          Manual, section 3.6.4.

          Async does not use a seperate SRC ISR.  Instead, the RCA
          ISR handles special receive conditions, so the RCA ISR address
          is placed in the SRC vector.  With async the RCA handles special
          receive conditions because it reads down the SCC's Rx FIFO, so each
          byte read must be checked for special conditions.  Reading down the
          Rx FIFO improves throughput by reducing the number of interrupts.

          Bisync uses seperate SRC and RCA ISR because the RCA ISR does NOT
          read down the SCC's Rx FIFO, as is done in async.  This is because
          Bisync must respond to framing characters quickly, so interrupts must
          be kept disabled for as short a time as possible.
Warnings: Disable interrupts before calling this function.
*/
void InitVector(void)
{
   // Table of SCC interrupt vectors.  Note that SCC A channels use odd
   // numbered indices in 1st array dimension, SCC B channels use even
   // numbered indices.
   static void (interrupt far *VectorTbl[NUMLINES][4])() =
   {
      {Ln1TbeIsr, Ln1EscIsr, Ln1RcaIsr, Ln1SrcIsr},
      {Ln0TbeIsr, Ln0EscIsr, Ln0RcaIsr, Ln0SrcIsr},
   };
   WORD_T v;                           // vector table index
   WORD_T i;
   WORD_T VectorType;                  // interrupt vector type number

#if LN0_HDLC                           // see "Comments:" above
   VectorTbl[1][2] = Ln0SrcIsr;
#endif
#if LN1_HDLC                           // see "Comments:" above
   VectorTbl[0][2] = Ln1SrcIsr;
#endif

#if LN0_ASYNC                          // see "Comments:" above
   VectorTbl[1][3] = Ln0RcaIsr;
#endif
#if LN1_ASYNC                          // see "Comments:" above
   VectorTbl[0][3] = Ln1RcaIsr;
#endif

   SetVector(SYS_TYPE,Sys2CtlIsr);     // initialize SYS to CTL ISR vector
   SetVector(DMATC_TYPE,DmaTcIsr);     // init DMA Terminal Cnt ISR vector

   // Initialize the vector table entries for each SCC
   VectorType = SCCBASE_TYPE;          // vector type for 1st SCC
   for(v = 0;v < NUMLINES;v++)         // for each vector table element (one
   {                                   //   each line)
      for(i = 0;i < 4;i++)             // there are 4 vectors for each line
      {
         SetVector(VectorType,VectorTbl[v][i]);
         VectorType += 2;             // skip unused vector
      }
   }
}

/***************************************************************************
Function: SetVector
Purpose:  Initialize an interrupt vector in the interrupt vector table.
Call:     SetVector(VectorType,Isr)
Entry:    WORD_T VectorType: Interrupt vector type number (0 - 0xff).
          void (interrupt far *Isr)(): Ptr to interrupt service routine.
Return:   void
Comments: The interrupt vector table is located in low memory, from 0:0 to
          0:3ff.  Each vector consists of a two byte offset followed by
          a two byte segment.  Together, this segment:offset pair forms
          a pointer to the interrupt service routine to be executed when
          that interrupt type occurs.  To get the starting address of a
          vector within the vector table the vector type is multiplied
          by 4.
Warnings: Disable interrupts before calling this function.
*/
void SetVector(WORD_T VectorType,void (interrupt far *Isr) ())
{
   UWORD_T far *vecttbl;               // ptr to vector table

   if(VectorType < 0 || VectorType > 0xff) // invalid vector type number
      return;

   vecttbl = (UWORD_T far *)(VectorType * 4);
   *vecttbl = FP_OFF(Isr);
   vecttbl++;
   *vecttbl = FP_SEG(Isr);
}

/***************************************************************************
Function: DispatchMsg
Purpose:  Remove all messages from the controller message queue and execute
          the indicated command functions.
Call:     DispatchCtlMsg()
Return:   void
Comments: The first byte of each system to controller message is the command.
          This it is an index into DispatchTbl[], it indicates the
          command function to be executed.

          This function is not reentrant.  Since it is possible for it to
          be called from either Sys2CtlIsr() or from the main dispatching loop
          reentrancy is blocked using the InDispatcher flag.
*/
void DispatchMsg(void)
{
   BYTE_T far *pMsg;                   // ptr to controller message
   BYTE_T Cmd;                         // command, is index into DispatchTbl[]

   // Dispatch table, one function for each command
   static void (*DispatchTbl[NUM_CTL_CMD])(BYTE_T far *) =
   {
      cmdTest,                         // 0 - CTLCMD_TEST
      cmdOpenAsync,                    // 1 - CTLCMD_OPENASYNC
      cmdOpenHdlc,                     // 2 - CTLCMD_OPENHDLC
      cmdOpenBisync,                   // 3 - CTLCMD_OPENBISYNC
      cmdClose,                        // 4 - CTLCMD_CLOSE
      cmdGetModemInputs,               // 5 - CTLCMD_GETMDM
      cmdSetModemOutputs               // 6 - CTLCMD_SETMDM
   };

   DIS_INT();
   if(InDispatcher)                    // already in this function
   {
      EN_INT();
      return;
   }
   InDispatcher = 1;
   EN_INT();

   // Process system to controller messages
   while(pCtlQ->In != pCtlQ->Out)      // while messages in queue
   {
      pMsg = pCtlQ->Buf[pCtlQ->Out];   // get ptr to message
      Cmd = *pMsg;                     // get command
      if(Cmd >= 0 && Cmd < NUM_CTL_CMD) // valid
         (*DispatchTbl[Cmd])(pMsg);    // execute command function
      pCtlQ->Out = (pCtlQ->Out + 1) % MSGQ_SIZE; // bump Out pointer
   }
   InDispatcher = 0;
}

/***************************************************************************
Function: EnqSysMsg
Purpose:  Add a message to the system message queue.
Call:     EnqSysMsg(pMsg)
          BYTE_T *pMsg; Ptr to message to be enqueued, must be MSG_LEN bytes
                        long.
Return:   1 if message successfully enqueued, else 0.
*/
WORD_T EnqSysMsg(BYTE_T *pMsg)
{
   WORD_T NumOpen;                        // number open message spaces in queue

   if((NumOpen = pSysQ->Out - pSysQ->In - 1) < 0) // get open space in queue
      NumOpen += MSGQ_SIZE;            // adjust for queue wrap
   if(NumOpen <= 0)                    // queue full
      return(0);                       // fail

   _fmemcpy(pSysQ->Buf[pSysQ->In],pMsg,MSG_LEN); // put in sys msg q
   pSysQ->In = (pSysQ->In + 1) % MSGQ_SIZE; // bump In pointer
   return(1);                          // success
}

/*****************************************************************************
Function: FarPtrToAbs
Purpose:  Convert a far pointer to an absolute address
Call:     FarPtrtoAbs(Ptr)
          void far *Ptr: The pointer to convert
Return:   UDWORD_T: The absolute address form of the pointer
*/
UDWORD_T FarPtrToAbs(void far *Ptr)
{
   return((((UDWORD_T)Ptr & 0xffff0000L) >> 12) +
          ((UDWORD_T)Ptr & 0x0000ffffL));
}

/****************************************************************************
Function: Timer1Isr
Purpose:  Timer 1 interrupt service routine.
          Increments heartbeat counter in firmware user area, which is used
          as the controller time counter, and detects changes in DSR and
          RI modem inputs.
Comments: There is no hardware to generate interrupts if DSR and RI change,
          so the change detection is handled here.  Note that the current
          state of CTS and CD is obtained from saved modem state rather that
          directly reading it from the SCC.  This is because the LnXEscIsr
          will update the saved copy if these change, so the saved copy is
          always up to date.
*/
void interrupt far Timer1Isr()
{
#if LN0_AYSNC || LN1_ASYNC
   UBYTE_T Cur;                        // current modem input state
   UBYTE_T Prev;                       // previous modem input state
   UBYTE_T Chg;                        // changed modem inputs
   WORD_T LnStatus;
#endif

   ISR_PROLOG();

   // Update controller time
   pFirm->heartbeat++;                 // bump heartbeat counter

#if LN0_AYSNC || LN1_ASYNC
   // Detect changes in DSR and RI for line 0.
   LnStatus = Line[0]->Status &
              (ASYNC_OPEN | HDLC_OPEN | BISYNC_OPEN | CLOSE_PEND);
   if(LnStatus & (ASYNC_OPEN | HDLC_OPEN | BISYNC_OPEN) &&
      !(LnStatus & CLOSE_PEND))
   {
      Prev = Line[0]->Modem;           // get previous modem state
      Cur = GET_DSR_RI(0) | (Prev & (CD | CTS)); // get current modem state
      Chg = Cur ^ Prev;                // get changed inputs
      if(Chg)                          // modem inputs changed
      {
         Line[0]->Modem = Cur;         // update modem state
#if TEST_MDMCHGMSG
         SysMsg[0] = SYSCMD_MDMCHG;    // msg to SYS, command field
         SysMsg[1] = 0;                //    line number field
         SysMsg[2] = Chg;              //    changed modem inputs field
         SysMsg[3] = Cur;              //    current modem inputs field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
#endif
      }
   }

   // Detect changes in DSR and RI for line 1.
   LnStatus = Line[1]->Status &
              (ASYNC_OPEN | HDLC_OPEN | BISYNC_OPEN | CLOSE_PEND);
   if(LnStatus & (ASYNC_OPEN | HDLC_OPEN | BISYNC_OPEN) &&
      !(LnStatus & CLOSE_PEND))
   {
      Prev = Line[1]->Modem;           // get previous modem state
      Cur = GET_DSR_RI(1) | (Prev & (CD | CTS)); // get current modem state
      Chg = Cur ^ Prev;                // get changed inputs
      if(Chg)                          // modem inputs changed
      {
         Line[1]->Modem = Cur;         // update modem state
#if TEST_MDMCHGMSG
         SysMsg[0] = SYSCMD_MDMCHG;    // msg to SYS, command field
         SysMsg[1] = 1;                //    line number field
         SysMsg[2] = Chg;              //    changed modem inputs field
         SysMsg[3] = Cur;              //    current modem inputs field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
#endif
      }
   }
#endif
   ISR_EPILOG();
}


/****************************************************************************
Function: Sys2CtlIsr
Purpose:  Process interrupt from system to controller.
Comments: This function dispatches controller messages.  In some applications
          it may not be desirable to do so at interrupt time because interrupts
          would be disabled for too long.  In such cases the main loop can do
          the dispatching instead of here.
*/
void interrupt far Sys2CtlIsr()
{
   ISR_PROLOG();
   DispatchMsg();
   ISR_EPILOG();
}

/***************************************************************************
Function: cmdTest
Purpose:  Controller command function to test operation of controller queue.
Call:     cmdTest(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_TEST
             Byte 1: Unused
             Byte 2: 0 = no reply, 1 = reply
Return:   void
*/
void cmdTest(BYTE_T far *pMsg)
{
   if(pMsg[2])
   {
      pCtlQ->ReplyBuf[0] = 'O';
      pCtlQ->ReplyBuf[1] = 'K';
      pCtlQ->ReplyBuf[2] = 0;
      pCtlQ->ReplyRdy = 1;
   }
}

/***************************************************************************
Function: cmdOpenAsync
Purpose:  Controller command function to open a line for asynchronous
          communications.
Call:     cmdOpenAsync(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_OPENASYNC
             Byte 1: Line number
             Byte 2: Baud rate parameter
             Byte 3: Parity parameter
             Byte 4: Data bits parameter
             Byte 5: Stop bits parameter
             Byte 6: Modem control parameter
Return:   void
*/
void cmdOpenAsync(BYTE_T far *pMsg)
{
#if LN0_ASYNC || LN1_ASYNC
   WORD_T LineNum;                     // line number
   LINE_T far *pLn;                    // ptr to line structure
   WORD_T Scc;                         // SCC command register address

   LineNum = pMsg[1];
   pLn = Line[LineNum];                // get ptr to line structure
   Scc = pLn->BaseIo;                  // get SCC command reg address
   pLn->Status = 0;                    // start out with line status cleared

   // Configure line structure using message passed from system
   pLn->WR3_ = RxDataBitTbl[pMsg[4]] | RX_ENABLE; // WR3, Rx char size
   pLn->WR4_ = StopBitTbl[pMsg[5]] |   // WR4, stop bits
               ParityTbl[pMsg[3]];     //      and parity
   pLn->WR5_ = TxDataBitTbl[pMsg[4]] | TX_ENABLE; // WR5, Tx char size
   pLn->WR12_ = BaudTbl[pMsg[2]];      // WR12, BRGT low byte
   pLn->WR13_ = BaudTbl[pMsg[2]] >> 8; // WR13, BRGT high byte
   pLn->WR15_ = CD_IE | CTS_IE;       // ESC interrupt enables

   // Initialize SCC from line structure
   DIS_INT();                          // protect SCC indexed accesses
   OUTB(Scc,WR4);                      // stop bits/parity + clock mode
   OUTB(Scc,pLn->WR4_ | X16_CLOCK);
   OUTB(Scc,WR2);                      // base interrupt vector type
   OUTB(Scc,pLn->WR2_);
   OUTB(Scc,WR3);                      // Rx character size
   OUTB(Scc,pLn->WR3_ & ~RX_ENABLE);
   OUTB(Scc,WR5);                      // Tx char size
   OUTB(Scc,pLn->WR5_ & ~TX_ENABLE);
   OUTB(Scc,WR9);                      // interrupt control
   OUTB(Scc,STATUS_LO | VIS);
   OUTB(Scc,WR11);                     // clock sources
   OUTB(Scc,RX_CLK_EQ_BRG | TX_CLK_EQ_BRG);
   OUTB(Scc,WR12);                     // lower byte of BRGTC
   OUTB(Scc,pLn->WR12_);
   OUTB(Scc,WR13);                     // upper byte of BRGTC
   OUTB(Scc,pLn->WR13_);
   OUTB(Scc,WR14);                     // BRG source
   OUTB(Scc,BRG_EQ_SYS_CLK | BRG_ENABLE);
   OUTB(Scc,WR3);                      // enable receive
   OUTB(Scc,pLn->WR3_);
   OUTB(Scc,WR5);                      // enable transmit
   OUTB(Scc,pLn->WR5_);
   EN_INT();

   // Initialize modem control outputs
   if(pMsg[6] & MSG_MDM_DTR)
      SetDtr(pLn);
   else
      ClrDtr(pLn);
   if(pMsg[6] & MSG_MDM_RTS)
   {
      SET_RTS(pLn);
   }
   else
   {
      CLR_RTS(pLn);
   }

   // Enable line
   pLn->TxqIn = 0;                     // flush the Tx queue
   pLn->TxqOut = 0;
   pLn->RxqIn = 0;                     // flush the Rx queue
   pLn->RxqOut = 0;
   DIS_INT();
   pLn->Status |= ASYNC_OPEN;          // indicate line open

   // Enable interrupts
   OUTB(Scc,WR15);                     // external/status interrupt control
   OUTB(Scc,pLn->WR15_);
   OUTB(Scc,RESET_EXT);                // WR0, reset external status interrupts
   OUTB(Scc,WR1);                      // interrupt enables
   OUTB(Scc,EXT_INT_ENABLE|PARITY_SPECIAL|RX_INT_ALL_SPECIAL);
   OUTB(Scc,WR9);                      // master interrupt enable
   OUTB(Scc,MIE+STATUS_LO+VIS);

   // Initialize saved state of modem inputs
   pLn->Modem = GET_CTS_CD(Scc);
   pLn->Modem |= GET_DSR_RI(LineNum);

   pCtlQ->ReplyRdy = 1;                // signal SYS open is done
   EN_INT();
#else
   pMsg = pMsg;
#endif
}

/***************************************************************************
Function: cmdOpenHdlc
Purpose:  Controller command function to open a line for HDLC communications.
Call:     cmdOpenHdlc(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_OPENHDLC
             Byte 1: Line number
             Byte 2: Modem control parameter
             Byte 3: Clock mode parameter
             Byte 4-7: Data rate in BPS, only used if Byte 3 = MSG_CLK_2.
                       The low order byte is in byte 4.
Return:   void
Comments: If the clock mode is MSG_CLK_1, both Tx and Rx clocks
          are programmed as inputs so they must be sourced by the DCE.

          If the clock mode is MSG_CLK_2 the Tx clock is programmed as an
          output sourced by the BRG, and the Rx clock as an input.  This is
          intended for use with a loopback plug that connects Tx clk to Rx
          clk and Tx to Rx - this is for test purposes.
*/
void cmdOpenHdlc(BYTE_T far *pMsg)
{
#if LN0_HDLC || LN1_HDLC
   LINE_T far *pLn;                    // ptr to line structure
   WORD_T Scc;                         // SCC command register address
   BYTE_T Wr11,Wr14;                   // WR11 & 14 contents
   WORD_T BRGen;                       // SCCbaud rate generator value
   WORD_T w;

   pLn = Line[pMsg[1]];                // get ptr to line structure
   Scc = pLn->BaseIo;                  // get SCC command reg address
   pLn->Status = 0;                    // start out with line status cleared

   // Configure line structure
   pLn->WR3_ = RX8_BIT_CHAR | RX_CRC_ENABLE | RX_ENABLE;
   pLn->WR4_ = SDLC_MODE;
   pLn->WR5_ = TX8_BIT_CHAR | TX_CRC_ENABLE | TX_ENABLE;
   pLn->WR15_ = CD_IE | CTS_IE | ABORT_IE; // ESC int enables

   // Initialize SCC
   DIS_INT();                          // protect SCC indexed accesses
   OUTB(Scc,WR4);                      // Tx/Rx mode
   OUTB(Scc,pLn->WR4_);
   OUTB(Scc,WR10);                     // Tx/Rx control
   OUTB(Scc,CRC_PRESET_1);
   OUTB(Scc,WR7);                      // SDLC flag
   OUTB(Scc,SDLC_FLAG);
   OUTB(Scc,WR3);                      // Rx parms
   OUTB(Scc,pLn->WR3_ & ~RX_ENABLE);
   OUTB(Scc,WR5);                      // Tx parms
   OUTB(Scc,pLn->WR5_ & ~TX_ENABLE);
   OUTB(Scc,WR2);                      // base interrupt vector type
   OUTB(Scc,pLn->WR2_);
   OUTB(Scc,WR9);                      // interrupt control
   OUTB(Scc,STATUS_LO+VIS);

   if(pMsg[3] == MSG_CLK_1)            // Tx & Rx clocks DCE source
   {
      ASM int CFGCTL_READINT
      ASM mov  w,ax
      w &= ~pLn->TxClkMask;
      ASM mov  ax,w;
      ASM int CFGCTL_WRITEINT
      Wr11 = TXCLK_TRxC | RXCLK_RTxC | TRxC_IN;
      Wr14 = BRG_DISABLE | TX_DMAREQ;
   }
   else if(pMsg[3] == MSG_CLK_2)       // Tx clk BRG DTE source, Rx clk DCE src
   {
      ASM int CFGCTL_READINT           // set up config ctrl reg
      ASM mov  w,ax
      w |= pLn->TxClkMask;             //    for Tx Clk DTE source
      ASM mov  ax,w
      ASM int CFGCTL_WRITEINT
      Wr11 = TX_CLK_EQ_BRG | TRxC_EQ_BRG | RXCLK_RTxC | TRxC_OUT;
      Wr14 = BRG_EQ_SYS_CLK | BRG_ENABLE | TX_DMAREQ;
      BRGen = (WORD_T)(((SCC_CLK / 2) / *(DWORD_T far *)&pMsg[4]) - 2);
      OUTB(Scc,WR12);                  // lower byte of BRGTC
      OUTB(Scc,BRGen);
      OUTB(Scc,WR13);                  // upper byte of BRGTC
      OUTB(Scc,BRGen >> 8);
   }
   OUTB(Scc,WR11);                     // clock sources
   OUTB(Scc,Wr11);
   OUTB(Scc,WR14);                     // BRG state, Tx DMA Req for DTR/REQ pin
   OUTB(Scc,Wr14);

   OUTB(Scc,WR5);                      // enable transmit
   OUTB(Scc,pLn->WR5_);
   OUTB(Scc,WR15);                     // allow write to WR7'
   OUTB(Scc,EN_WR7PRIME);
   OUTB(Scc,WR7);                      // SDLC enhancements
   OUTB(Scc,AUTOTX_FLAG | AUTOEOM_RESET | RXCRC);
   EN_INT();

   // Initialize modem control outputs
   if(pMsg[2] & MSG_MDM_DTR)
      SetDtr(pLn);
   else
      ClrDtr(pLn);
   if(pMsg[2] & MSG_MDM_RTS)
   {
      SET_RTS(pLn);
   }
   else
   {
      CLR_RTS(pLn);
   }

   // Enable line
   pLn->TxbqIn = 0;                    // flush the Tx buffer queue
   pLn->TxbqOut = 0;
   pLn->TxbqCnt = 0;
   pLn->TxbqRetryCnt = 0;
   pLn->RxbqIn = 0;                    // flush the Rx buffer queue
   pLn->RxbqOut = 0;
   pLn->RxbqCnt = 0;
   DIS_INT();
   pLn->Status |= HDLC_OPEN;           // indicate line active

   // Enable interrupts
   OUTB(Scc,WR15);                     // external/status interrupts
   OUTB(Scc,pLn->WR15_);
   OUTB(Scc,RESET_EXT);                // WR0, reset external status interrupts
   OUTB(Scc,WR1);                      // interrupt enables, Rx DMA Req
   OUTB(Scc,EXT_INT_ENABLE | RX_INT_SPECIAL | RX_DMA_REQ);
   OUTB(Scc,WR1);                      // enable Rx DMA
   OUTB(Scc,EXT_INT_ENABLE | RX_INT_SPECIAL | RX_DMA_REQ | RX_DMA_ENABLE);
   OUTB(Scc,WR9);                      // master interrupt enable
   OUTB(Scc,MIE+STATUS_LO+VIS);
   OUTB(INTCTL1,INB(INTCTL1) & 0xfe);  // enable DMA terminal count interrupt
   pLn->RxState = HRX_BUF_WAIT;        // init Rx state machine
   pLn->TxState = STX_IDLE;            // init Tx state machine

   // Initialize saved state of modem inputs
   pLn->Modem = GET_CTS_CD(Scc);
   pLn->Modem |= GET_DSR_RI(pMsg[1]);

   pCtlQ->ReplyRdy = 1;                // signal SYS open is done
   EN_INT();
#else
   pMsg = pMsg;
#endif
}

/***************************************************************************
Function: cmdOpenBisync
Purpose:  Controller command function to open a line for Bisync communications.
Call:     cmdOpenBisync(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_OPENBISYNC
             Byte 1: Line number
             Byte 2: Modem control parameter
             Byte 3: Clock mode parameter
Return:   void
Comments: If the clock mode is MSG_CLK_1, both Tx and Rx clocks
          are programmed as inputs so they must be sourced by the DCE.

          If the clock mode is MSG_CLK_2 the Tx clock is programmed as an
          output sourced by the BRG, and the Rx clock as an input.  This is
          intended for use with a loopback plug that connects Tx clk to Rx
          clk and Tx to Rx - this is for test purposes.
*/
void cmdOpenBisync(BYTE_T far *pMsg)
{
#if LN0_BISYNC || LN1_BISYNC
   LINE_T far *pLn;                    // ptr to line structure
   WORD_T Scc;                         // SCC command register address
   BYTE_T Wr11,Wr14;                   // WR11 & 14 contents
   WORD_T BRGen;                       // SCCbaud rate generator value
   WORD_T w;

   pLn = Line[pMsg[1]];                // get ptr to line structure
   Scc = pLn->BaseIo;                  // get SCC command reg address
   pLn->Status = 0;                    // start out with line status cleared

   // Configure line structure
   pLn->WR3_ = RX8_BIT_CHAR | RX_ENABLE;
   pLn->WR4_ = BISYNC_MODE;
   pLn->WR5_ = TX8_BIT_CHAR | TX_ENABLE | CRC_16 | TX_CRC_ENABLE;
   pLn->WR15_ = CD_IE | CTS_IE;        // ESC interrupt enables

   // Initialize SCC
   DIS_INT();                          // protect SCC indexed accesses
   OUTB(Scc,WR4);                      // Tx/Rx mode
   OUTB(Scc,pLn->WR4_);
   OUTB(Scc,WR10);                     // Tx/Rx control
   OUTB(Scc,CRC_PRESET_1);
   OUTB(Scc,WR6);                      // sync char
   OUTB(Scc,SYN);
   OUTB(Scc,WR7);                      // sync char
   OUTB(Scc,SYN);
   OUTB(Scc,WR3);                      // Rx parms
   OUTB(Scc,pLn->WR3_ & ~RX_ENABLE);
   OUTB(Scc,WR5);                      // Tx parms
   OUTB(Scc,pLn->WR5_ & ~TX_ENABLE);
   OUTB(Scc,WR2);                      // base interrupt vector type
   OUTB(Scc,pLn->WR2_);
   OUTB(Scc,WR9);                      // interrupt control
   OUTB(Scc,STATUS_LO+VIS);
   if(pMsg[3] == MSG_CLK_1)            // Tx & Rx clocks DCE source
   {
      ASM int CFGCTL_READINT
      ASM mov  w,ax
      w &= ~pLn->TxClkMask;
      ASM mov  ax,w;
      ASM int CFGCTL_WRITEINT
      Wr11 = TXCLK_TRxC | RXCLK_RTxC | TRxC_IN;
      Wr14 = BRG_DISABLE | TX_DMAREQ;
   }
   else if(pMsg[3] == MSG_CLK_2)       // Tx clk BRG DTE source, Rx clk DCE src
   {
      ASM int CFGCTL_READINT           // set up config ctrl reg
      ASM mov  w,ax
      w |= pLn->TxClkMask;             //    for Tx Clk DTE source
      ASM mov  ax,w
      ASM int CFGCTL_WRITEINT
      Wr11 = TX_CLK_EQ_BRG | TRxC_EQ_BRG | RXCLK_RTxC | TRxC_OUT;
      Wr14 = BRG_EQ_SYS_CLK | BRG_ENABLE | TX_DMAREQ;
      BRGen = (WORD_T)(((SCC_CLK / 2) / *(DWORD_T far *)&pMsg[4]) - 2);
      OUTB(Scc,WR12);                  // lower byte of BRGTC
      OUTB(Scc,BRGen);
      OUTB(Scc,WR13);                  // upper byte of BRGTC
      OUTB(Scc,BRGen >> 8);
   }
   OUTB(Scc,WR11);                     // clock sources
   OUTB(Scc,Wr11);
   OUTB(Scc,WR14);                     // BRG state, Tx DMA Req for DTR/REQ pin
   OUTB(Scc,Wr14);

   OUTB(Scc,WR5);                      // enable transmit
   OUTB(Scc,pLn->WR5_);
   EN_INT();

   // Initialize modem control outputs
   if(pMsg[2] & MSG_MDM_DTR)
      SetDtr(pLn);
   else
      ClrDtr(pLn);
   if(pMsg[2] & MSG_MDM_RTS)
   {
      SET_RTS(pLn);
   }
   else
   {
      CLR_RTS(pLn);
   }

   // Enable line
   pLn->TxbqIn = 0;                    // flush the Tx buffer queue
   pLn->TxbqOut = 0;
   pLn->TxbqCnt = 0;
   pLn->TxbqRetryCnt = 0;
   pLn->RxbqIn = 0;                    // flush the Rx buffer queue
   pLn->RxbqOut = 0;
   pLn->RxbqCnt = 0;
   pLn->RxbqCurCnt = 0;
   DIS_INT();
   pLn->Status |= BISYNC_OPEN;         // indicate line active

   // Enable interrupts
   OUTB(Scc,WR15);                     // external/status interrupts
   OUTB(Scc,pLn->WR15_);
   OUTB(Scc,RESET_EXT);                // WR0, reset external status interrupts
   OUTB(Scc,WR1);                      // interrupt enables
   OUTB(Scc,EXT_INT_ENABLE|PARITY_SPECIAL|RX_INT_ALL_SPECIAL);
   OUTB(Scc,WR9);                      // master interrupt enable
   OUTB(Scc,MIE+STATUS_LO+VIS);
   OUTB(INTCTL1,INB(INTCTL1) & 0xfe);  // enable DMA terminal count interrupt
   pLn->RxState = BRX_BUF_WAIT;        // init Rx state machine
   pLn->TxState = STX_IDLE;            // init Tx state machine

   // Initialize saved state of modem inputs
   pLn->Modem = GET_CTS_CD(Scc);
   pLn->Modem |= GET_DSR_RI(pMsg[1]);

   pCtlQ->ReplyRdy = 1;                // signal SYS open is done
   EN_INT();
#else
   pMsg = pMsg;
#endif
}

/***************************************************************************
Function: cmdClose
Purpose:  Controller command function to close a line
Call:     cmdClose(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_CLOSE
             Byte 1: Line number
Return:   void
Comments: This function does not immediately close the line, instead it
          marks the line as close pending.  This is done so the main
          dispatching loop can allow Tx buffers to empty before
          closing.
*/
void cmdClose(BYTE_T far *pMsg)
{
   LINE_T far *pLn;                    // ptr to line structure
   UDWORD_T Timeout;                   // controller timeout time
   UDWORD_T Time;                      // current controller time

   pLn = Line[pMsg[1]];
   pLn->Status |= CLOSE_PEND;
   Time = GetCtlTime();
   if(pLn->Status & ASYNC_OPEN)
      Timeout = Time + MAX_CLOSE_DLY;  // how long to wait for Tx q to empty
   else
      Timeout = Time + SYNC_CLOSE_DLY; // how long to delay to allow SYNC to Tx
   if(Timeout < Time)                  // CTL time will roll over
   {                                   // force the rollover now
      DIS_INT();                       // block timer int between word accesses
      pFirm->heartbeat = 0L;
      Timeout = MAX_CLOSE_DLY;
      EN_INT();
   }
   pLn->Timeout = Timeout;
}

/***************************************************************************
Function: Close
Purpose:  Close a line.
Call:     Close(pLn)
          LINE_T far *pLn: Ptr to a line structure
Return:   void
*/
void Close(LINE_T far *pLn)
{
   WORD_T Scc;                         // SCC command register address
   UBYTE_T b;                          // data byte

   Scc = pLn->BaseIo;                  // get SCC command reg address

#if LN0_ASYNC || LN1_ASYNC
   if(pLn->Status & ASYNC_OPEN)        // async close
   {
      b = INB(Scc+2);                     // clear SCC Rx buffer
      b = INB(Scc+2);
      b = INB(Scc+2);
      if(pLn->LineNum & 1)                // odd numbered line
         b = STATUS_LO+MIE+VIS+RESET_CH_B; // reset value for SCC channel B
      else                                // even numbered line
         b = STATUS_LO+MIE+VIS+RESET_CH_A; // reset value for SCC channel A
      DIS_INT();                          // protect SCC indexed accesses
      OUTB(Scc,WR9);                      // reset SCC channel
      OUTB(Scc,b);
      ClrDtr(pLn);                        // turn DTR OFF
      pLn->TxqOut = 0;                    // flush Tx and Rx queues
      pLn->TxqIn = 0;
      pLn->RxqOut = 0;
      pLn->RxqIn = 0;
      pLn->Status = 0;                    // disable all bits in Status
      EN_INT();
   }
#endif

#if LN0_BISYNC || LN1_BISYNC || LN0_HDLC || LN1_HDLC
   if(pLn->Status & (HDLC_OPEN | BISYNC_OPEN)) // sync close
   {
      DIS_INT();
      DIS_DMA(pLn->DisTxDma | pLn->DisRxDma); // disable DMA
      b = INB(Scc+2);                     // clear SCC Rx buffer
      b = INB(Scc+2);                     // clear SCC Rx buffer
      b = INB(Scc+2);                     // clear SCC Rx buffer
      if(pLn->LineNum & 1)                // odd numbered line
        b = STATUS_LO+MIE+VIS+RESET_CH_B; // reset value for SCC channel B
      else                                // even numbered line
         b = STATUS_LO+MIE+VIS+RESET_CH_A; // reset value for SCC channel A
      OUTB(Scc,WR9);                      // reset SCC channel
      OUTB(Scc,b);
      ClrDtr(pLn);                        // turn DTR OFF
      pLn->TxbqOut = 0;                   // flush Tx and Rx buffer queues
      pLn->TxbqIn = 0;
      pLn->TxbqCnt = 0;
      pLn->TxbqRetryCnt = 0;
      pLn->RxbqOut = 0;
      pLn->RxbqIn = 0;
      pLn->RxbqCnt = 0;
      pLn->Status = 0;                    // disable all bits in Status
      pLn->TxState = 0;                   // clear state machines
      pLn->RxState = 0;
      EN_INT();
   }
#endif
}

/***************************************************************************
Function: cmdGetModemInputs
Purpose:  Controller command function to get modem control inputs.
Call:     cmdGetModemInputs(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_GETMDM
             Byte 1: Line number
Return:   void. Current state of modem control inputs is returned to the
          system in a one byte reply.  The reply sets a MSG_MDM_X flag
          for each modem input that is ON.
Comments: A copy of the current state of the modem inputs is saved in the
          line structure.  The return value is obtained from there after
          updating it for DSR and RI.  The update in saved modem state is
          done here to avoid the Timer1Isr() latency in detecting changes on
          DSR and RI.
*/
void cmdGetModemInputs(BYTE_T far *pMsg)
{
   WORD_T Ln;                          // line number
   UBYTE_T Cur, Prev;                  // current and previous modem states

   Ln = pMsg[1];
   DIS_INT();
   Prev = Line[Ln]->Modem;             // get previous modem state
   Cur = GET_DSR_RI(Ln) | (Prev & (CD | CTS)); // get current modem state
   Line[Ln]->Modem = Cur;              // update modem state
   EN_INT();
   pCtlQ->ReplyBuf[0] = Cur;           // put current state in reply
   pCtlQ->ReplyRdy = 1;                // indicate reply ready
}

/***************************************************************************
Function: cmdSetModemOutputs
Purpose:  Controller command function to set state of modem control inputs.
Call:     cmdGetModemInputs(pMsg)
          BYTE_T far *pMsg; Ptr to message passed from system. Format is:
             Byte 0: Command, CTLCMD_GETMDM
             Byte 1: Line number
             Byte 2: RTS state, must be one of flags ON, OFF, or NOCHANGE
             Byte 3: DTR state, must be one of flags ON, OFF, or NOCHANGE
Return:   void
*/
void cmdSetModemOutputs(BYTE_T far *pMsg)
{
   LINE_T far *pLn;                    /* ptr to line descriptor */

   pLn = Line[pMsg[1]];
   if(pMsg[2] == OFF)                  /* turn RTS OFF */
   {
      CLR_RTS(pLn);
   }
   else if(pMsg[2] == ON)              /* turn RTS ON */
   {
      SET_RTS(pLn);
   }

   if(pMsg[3] == OFF)                  /* turn DTR OFF */
   {
      ClrDtr(pLn);
   }
   else if(pMsg[3] == ON)              /* turn DTR ON */
   {
      SetDtr(pLn);
   }
}

/***************************************************************************
Function: SetDtr
Purpose:  Set DTR modem control output
Call:     SetDtr(pLn)
          LINE_T far *pLn: Ptr to line structure
Return:   void
Comments: This is used to set the DTR modem control output via the
          configuration control register rather than the SCC.  This is done
          to free up the SCC DTR/REQ pin so it can be used as the Tx DMA
          request pin.
*/
void SetDtr(LINE_T far *pLn)
{
   WORD_T w;

   DIS_INT();
   ASM int CFGCTL_READINT
   ASM mov  w,ax
   w &= ~pLn->AltDtrMask;
   ASM mov  ax,w
   ASM int CFGCTL_WRITEINT
   EN_INT();
}

/***************************************************************************
Function: ClrDtr
Purpose:  Clear DTR modem control output
Call:     ClrDtr(pLn)
          LINE_T far *pLn: Ptr to line structure
Return:   void
Comments: This is used to clear the DTR modem control output via the
          configuration control register rather than the SCC.  This is done
          to free up the SCC DTR/REQ pin so it can be used as the Tx DMA
          request pin.
*/
void ClrDtr(LINE_T far *pLn)
{
   WORD_T w;

   DIS_INT();
   ASM int CFGCTL_READINT
   ASM mov  w,ax
   w |= pLn->AltDtrMask;
   ASM mov  ax,w
   ASM int CFGCTL_WRITEINT
   EN_INT();
}

/***************************************************************************
Function: RetryTxHdlc
Purpose:  Retry sending an aborted HDLC frame.
Call:     RetryTxHdlc(pLn);
          LINE_T far *pLn;             // ptr to line structure
Return:   void
*/
void RetryTxHdlc(LINE_T far *pLn)
{
   static BYTE_T Msg[MSG_LEN] =        // SYS msg buffer
   {
      SYSCMD_TXERROR,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
   };

   DIS_INT();
   if(++pLn->TxbqRetryCnt > TXRETRY_MAX) // exceeded max retries
   {
      Trace(pLn->LineNum | 0xc000,0,0,0);
      pLn->TxState = STX_CLOSE_WAIT;
      Msg[1] = pLn->LineNum;           // SYS msg line number field
      Msg[2] = MSG_TXERR_RETRYMAX;     //   error flag field
      EnqSysMsg(Msg);                  // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
   }
   else                                // OK to retry
   {
      Trace(pLn->LineNum | 0xc100,pLn->TxbqRetryCnt,0,0);
      pLn->TxState = STX_IDLE;         // state machine to Tx idle
   }
   EN_INT();
}

#if LN0_HDLC || LN1_HDLC || LN0_BISYNC || LN1_BISYNC
/***************************************************************************
Function: StartTxSync
Purpose:  Start up synchronous transmit by initializing the DMA channel.
Call:     StartTxSync(pLn);
          LINE_T far *pLn;             // ptr to line structure
Return:   void
Comments: This is used for both HDLC and BISYNC.
Warnings: Don't call this function unless a Tx buffer is available.
          No checking for this is done.
*/
void StartTxSync(LINE_T far *pLn)
{
   BUFFER_T far *pBuf;                 // ptr to buffer structure

   pBuf = &pLn->Txb[pLn->TxbqOut];     // get ptr to buffer struct
   if(pBuf->Cnt > 0)                   // there is data in the buffer
   {
      // Configure V53 DMA channel for transmit data
      DIS_INT();
      OUTB(DCH,pLn->TxDmaChan);        // set up to access Tx dma channel
      OUTW(DCA,pBuf->BufLo);           // Tx buffer absolute low word
      OUTB(DCA+2,pBuf->BufHi);         // Tx buffer absolute high word
      OUTW(DCC,pBuf->Cnt - 1);         // xfer count, xfers 1 more than DCC

      // Enable Tx DMA channel, this starts the Tx DMA
      EN_DMA(pLn->EnTxDma);
      pLn->TxState = STX_ACTIVE;       // state machine to Tx DMA active state

      // By now the first Tx byte will have been written to the SCC, so OK
      // to do SCC operations that must be done after writing the first byte
      OUTB(pLn->BaseIo,RESET_TXUNDERRUN); // WR0, reset Tx underrun latch
      pLn->WR15_ |= TXUNDERRUN_IE;     // enable Tx underrun ints
      OUTB(pLn->BaseIo,WR15);
      OUTB(pLn->BaseIo,pLn->WR15_);
      EN_INT();
   }
   else                                // the buffer is empty
   {
      pLn->TxState = STX_EOF;          // state machine to EOF state
   }
}
#endif


#if LN0_HDLC || LN1_HDLC
/***************************************************************************
Function: StartRxHdlc
Purpose:  Start up HDLC receive by initializing the DMA channel and enabling
          the receiver.
Call:     StartRxHdlc(pLn);
          LINE_T far *pLn;             // ptr to line structure
Return:   void
Warnings: Don't call this function unless an Rx buffer is available.
          No checking for this is done.
*/
void StartRxHdlc(LINE_T far *pLn)
{
   BUFFER_T far *pBuf;                 // ptr to buffer structure

   // Configure V53 DMA channel for receive data
   pBuf = &pLn->Rxb[pLn->RxbqIn];      // get ptr to buffer struct
   DIS_INT();
   OUTB(DCH,pLn->RxDmaChan);           // set up to access Rx dma channel
   OUTW(DCA,pBuf->BufLo);              // Rx buffer absolute low word
   OUTB(DCA+2,pBuf->BufHi);            // Rx buffer absolute high word
   OUTW(DCC,RXBQ_BUF_SIZE - 1);        // xfer cnt, xfers 1 more than DCC

   // Enable SCC receive, put it in Hunt mode so won't get anything
   // until SYNC chars are received.
   OUTB(pLn->BaseIo,WR3);
   OUTB(pLn->BaseIo,pLn->WR3_ | ENTER_HUNT_MODE);

   // Enable Rx DMA channel, this starts the Rx DMA if data is available
   EN_DMA(pLn->EnRxDma);
   pLn->RxState = HRX_ACTIVE;          // state machine to Rx DMA active state
   EN_INT();
}
#endif

#if LN0_BISYNC || LN1_BISYNC
/***************************************************************************
Function: StartRxBisync
Purpose:  Start up Bisync receive
Call:     StartRxBisync(pLn);
          LINE_T far *pLn;             // ptr to line structure
Return:   void
Warnings: Don't call this function unless an Rx buffer is available.
          No checking for this is done.
*/
void StartRxBisync(LINE_T far *pLn)
{
   // Initialize current Rxbq buffer ptr and count
   DIS_INT();
   pLn->RxbqCurBuf = pLn->Rxb[pLn->RxbqIn].BufCtl;
   pLn->RxbqCurCnt = 0;

   // Enable SCC receive, put it in Hunt mode so won't get anything
   // until SYN chars are received.
   OUTB(pLn->BaseIo,WR3);
   OUTB(pLn->BaseIo,pLn->WR3_ | ENTER_HUNT_MODE);
   pLn->RxState = BRX_SOH_WAIT;        // wait for SOH character state
   EN_INT();
}
#endif

/***************************************************************************
Function: InitDmaCtrl
Purpose:  Initialize DMA channels 0 through 3
Call:     InitDmaCtrl()
Return:   void
Comments: All four DMA channels are initialized but left with DMA requests
          masked off.  This prevents any DMA from taking place until the
          line that will do DMA is initialized, at that time the proper
          channels are masked on.  DMA channels 0 and 1 are set up for Tx
          and Rx respectively on line 0, channels 2 and 3 for Tx and Rx
          respectively on line 1.
*/
void InitDmaCtrl(void)
{
   OUTB(DICM,1);           // reset DMA controller

   OUTB(DCH,0);            // set up to access channel 0
   OUTB(DMD,0x48);         // DMA mode control register,single transfer mode,
                           // increment address, disable autoinitialize, memory
                           // to I/O, byte

   OUTB(DCH,1);            // set up to access channel 1
   OUTB(DMD,0x44);         // DMA mode control register,single transfer mode,
                           // increment address, disable autoinitialize, I/O to
                           // memory, byte

   OUTB(DCH,2);            // set up to access channel 2
   OUTB(DMD,0x48);         // DMA mode control register,single transfer mode,
                           // increment address, disable autoinitialize, memory
                           // to I/O, byte

   OUTB(DCH,3);            // set up to access channel 3
   OUTB(DMD,0x44);         // DMA mode control register,single transfer mode,
                           // increment address, disable autoinitialize, I/O to
                           // memory, byte

   OUTW(DDC,0);            // DMA device control register, normal mode, fixed
                           // priority, enabled DMA's, wait during verify
                           // disabled, bus rel mode

   OUTB(DMK,DMAMASK);      // mask DMA request off for now
}

/****************************************************************************
Function: DmaTcIsr
Purpose:  Process interrupt for DMA terminal count
Comments: This ISR checks the DST (DMA Status Register) to see which DMA
          channels have TC (Terminal Count).  When the DST is read it is
          cleared.  This means that all DMA channels with TC set must have
          that TC processed within the ISR, otherwise that TC will be lost.
*/
void interrupt far DmaTcIsr()
{
   BYTE_T Status;                      // DST (DMA Status) register contents
   LINE_T far *pLn;                    // ptr to line structure
   BUFFER_T far *pBuf;                 // ptr to buffer structure

   ISR_PROLOG();
   Status = INB(DST);                  // read DST

   if(Status & TC0)                    // terminal count on DMA channel 0
   {
      // This TC indicates that the last Tx byte has been written to line 0.
      // We now expect a Tx underrun to occur after the last byte of the
      // frame has shifted out.  When this happens we want the CRC and a flag
      // to be transmitted rather than an abort, so disable abort underrun.
      pLn = Line[0];
      pLn->TxbqRetryCnt = 0;       // frame sent OK, clear retry count
      pLn->TxState = STX_EOF_WAIT; // state machine to wait for Tx EOF
   }

   if(Status & TC1)                    // terminal count on DMA channel 1
   {
      // This TC indicates line 0 receive DMA is about to exceed it's Rx
      // buffer size.  Put SCC back in Hunt mode so that no more data will
      // be received until a new frame starts.  Restart DMA on same buffer.
      // This frame will be discarded, the next frame will be placed in the
      // same buffer this frame was going in.
      pLn = Line[0];
      OUTB(pLn->BaseIo,WR3);           // make SCC look for next frame
      OUTB(pLn->BaseIo,pLn->WR3_ & ENTER_HUNT_MODE);

      pBuf = &pLn->Rxb[pLn->RxbqIn];   // get ptr to buffer struct
      OUTB(DCH,LN0_RXDMACH);           // set up to access Rx dma channel
      OUTW(DCA,pBuf->BufLo);           // Rx buffer absolute low word
      OUTB(DCA+2,pBuf->BufHi);         // Rx buffer absolute high word
      OUTW(DCC,RXBQ_BUF_SIZE - 1);     // transfer cnt, xfers 1 more than DCC
      EN_DMA(LN0_ENRXDMA);             // enable Rx DMA channel

      SysMsg[0] = SYSCMD_RXERROR;      // messge to SYS, cmd field
      SysMsg[1] = 0;                   //   line number field
      SysMsg[2] = MSG_RXERR_OVERFLOW;  //   error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc200,0,0,0);
   }

   if(Status & TC2)                    // terminal count on DMA channel 2
   {
      // This TC indicates that the last Tx byte has been written to line 1.
      // We now expect a Tx underrun to occur after the last byte of the
      // frame has shifted out.  When this happens we want the CRC and a flag
      // to be transmitted rather than an abort, so disable abort underrun.
      pLn = Line[1];
      pLn->TxbqRetryCnt = 0;       // frame sent OK, clear retry count
      pLn->TxState = STX_EOF_WAIT; // state machine to wait for Tx EOF
   }

   if(Status & TC3)                    // terminal count on DMA channel 3
   {
      // This TC indicates line 1 receive DMA is about to exceed it's Rx
      // buffer size.  Disable the receiver and put state machine back to
      // waiting for a Rx buffer.  This has effect of putting SCC back in Hunt
      // mode so that no more data will be received until a new frame starts.
      // This frame will be discarded, the next frame will be placed in the
      // same buffer this frame was going in.
      pLn = Line[1];
      OUTB(pLn->BaseIo,WR3);           // make SCC look for next frame
      OUTB(pLn->BaseIo,pLn->WR3_ & ENTER_HUNT_MODE);

      pBuf = &pLn->Rxb[pLn->RxbqIn];   // get ptr to buffer struct
      OUTB(DCH,LN1_RXDMACH);           // set up to access Rx dma channel
      OUTW(DCA,pBuf->BufLo);           // Rx buffer absolute low word
      OUTB(DCA+2,pBuf->BufHi);         // Rx buffer absolute high word
      OUTW(DCC,RXBQ_BUF_SIZE - 1);     // transfer cnt, xfers 1 more than DCC
      EN_DMA(LN1_ENRXDMA);             // enable Rx DMA channel

      SysMsg[0] = SYSCMD_RXERROR;      // messge to SYS, cmd field
      SysMsg[1] = 1;                   //   line number field
      SysMsg[2] = MSG_RXERR_OVERFLOW;  //   error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc201,0,0,0);
   }

   ISR_EPILOG();
}

/****************************************************************************
Function: LnXTbeIsr
Purpose:  Transmit Buffer Empty Interrupt Service Routines, one for each
          line.
Comments: Since none of the communication modes not use interrupt driven
          transmit, the async ISRs do no work and should never be called.
*/
#if LN0_ASYNC
void interrupt far Ln0TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_ASYNC
void interrupt far Ln1TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_HDLC
void interrupt far Ln0TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_HDLC
void interrupt far Ln1TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_BISYNC
void interrupt far Ln0TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_BISYNC
void interrupt far Ln1TbeIsr()
{
   ISR_PROLOG();
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

/*****************************************************************************
Function: LnXEscIsr
Purpose:  External Status Change Interrupt Service Routines, one for each
          line.
Comments: To achieve reliable 64k bps operation on 2 ports it was necessary
          to partially code the HDLC ISRs in assembly.  Further
          performance gains could by made by rewriting the entire ISR in
          assembly.
*/
#if LN0_ASYNC
void interrupt far Ln0EscIsr()
{
   LINE_T far *pLn;
   UBYTE_T Cur;                        // current modem input state
   UBYTE_T Prev;                       // previous modem input state
   UBYTE_T Chg;                        // changes modem inputs

   ISR_PROLOG();
   pLn  = Line[0];
   Cur = INB(LN00_IO) & (CD | CTS);    // get current modem state
   Cur |= GET_DSR_RI(0);               //    including DSR-RI inputs
   Prev = pLn->Modem;                  // get previous modem state
   Chg = Cur ^ Prev;                   // get changed inputs
   if(Chg)                             // modem inputs changed
   {
      pLn->Modem = Cur;                // update modem state
#if TEST_MDMCHGMSG
      SysMsg[0] = SYSCMD_MDMCHG;       // msg to SYS, command field
      SysMsg[1] = 0;                   //    line number field
      SysMsg[2] = Chg;                 //    changed modem inputs field
      SysMsg[3] = Cur;                 //    current modem inputs field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
#endif
   }

   OUTB(LN00_IO,RESET_EXT);            // WR0, reset external status interrupts
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_ASYNC
void interrupt far Ln1EscIsr()
{
   LINE_T far *pLn;
   UBYTE_T Cur;                        // current modem input state
   UBYTE_T Prev;                       // previous modem input state
   UBYTE_T Chg;                        // changes modem inputs

   ISR_PROLOG();
   pLn  = Line[1];
   Cur = INB(LN01_IO) & (CD | CTS);        // get current modem state
   Cur |= GET_DSR_RI(1);               //    including DSR-RI inputs
   Prev = pLn->Modem;                  // get previous modem state
   Chg = Cur ^ Prev;                   // get changed inputs
   if(Chg)                             // modem inputs changed
   {
      pLn->Modem = Cur;                // update modem state
#if TEST_MDMCHGMSG
      SysMsg[0] = SYSCMD_MDMCHG;       // msg to SYS, command field
      SysMsg[1] = 1;                   //    line number field
      SysMsg[2] = Chg;                 //    changed modem inputs field
      SysMsg[3] = Cur;                 //    current modem inputs field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
#endif
   }

   OUTB(LN01_IO,RESET_EXT);                // WR0, reset external status interrupts
   OUTB(LN01_IO,RESET_IUS);                // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_HDLC
void interrupt far Ln0EscIsr()
{
   BUFFER_T far *pBuf;                 // ptr to buffer structure
   LINE_T far *pLn;
   UBYTE_T Status;                     // SCC Tx, Rx, and external status
   WORD_T TxState;                     // current state of Tx state machine
   BYTE_T b;


   ISR_PROLOG();
   pLn  = Line[0];
   Status = INB(LN00_IO) & pLn->WR15_; // read Ext status, enabled ESC's only

   if(Status & TXUNDERRUN_COND)        // a Tx underrun occurred
   {
      TxState = pLn->TxState;          // Tx state machine, current state
      if(TxState == STX_EOF_WAIT)      // valid end of Tx frame
      {
         // Disable Tx underrun interrups
         pLn->WR15_ &= ~TXUNDERRUN_IE;
         b = pLn->WR15_;
         ASM {
            push dx
            mov  dx,LN00_IO
            mov  al,WR15
            out  dx,al
            mov  al,b
            out  dx,al
            pop  dx
         }
         pLn->TxState = STX_EOF;       // state machine to Tx EOF processing
      }
      else                             // aborted Tx frame
      {
         DIS_DMA(LN0_DISTXDMA);        // disable Tx DMA channel
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN00_IO,WR15);
         OUTB(LN00_IO,pLn->WR15_);
         OUTB(LN00_IO,SEND_ABORT);     // send an abort
         Trace(0xc300,pLn->TxbqOut,TxState,pLn->TxbqRetryCnt);
         pLn->TxState = STX_RETRY;     // state machine to Tx retry state
      }
      OUTB(LN00_IO,RESET_TXUNDERRUN);  // reset Tx underrun latch
   }

   if(Status & ABORT_COND)             // Rx abort sequence occurred
   {
      // This frame will be discarded, re-initialize DMA so the next frame
      // will be placed in the same buffer this frame was going in.  Put
      // SCC back in Hunt mode so that no more data will be received until
      // a new frame starts.
      pBuf = &pLn->Rxb[pLn->RxbqIn];   // get ptr to buffer struct
      OUTB(DCH,LN0_RXDMACH);           // set up to access Rx dma channel
      OUTW(DCA,pBuf->BufLo);           // Rx buffer absolute low word
      OUTB(DCA+2,pBuf->BufHi);         // Rx buffer absolute high word
      OUTW(DCC,RXBQ_BUF_SIZE - 1);     // xfer cnt, xfers 1 more than DCC
      OUTB(LN00_IO,WR3);               // put receiver in hunt mode
      OUTB(LN00_IO,pLn->WR3_ | ENTER_HUNT_MODE);

      SysMsg[0] = SYSCMD_RXERROR;      // message to SYS, cmd field
      SysMsg[1] = 0;                   //    line number field
      SysMsg[2] = MSG_RXERR_ABORT;     //    error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc400,pLn->RxbqIn,0,0);
   }

   // SCC: reset external status interrrupts and end of interrupt
   // PIC: end of interrupt
   ASM {
      mov  dx,LN00_IO
      mov  al,RESET_EXT
      out  dx,al
      mov  al,RESET_IUS
      out  dx,al
      mov  InIsr,0
      mov  dx,INTCTL
      mov  al,EOIVAL
      out  dx,al
   }
}
#endif // #if LN0_HDLC

#if LN1_HDLC
void interrupt far Ln1EscIsr()
{
   BUFFER_T far *pBuf;                 // ptr to buffer structure
   LINE_T far *pLn;
   UBYTE_T Status;                     // SCC Tx, Rx, and external status
   WORD_T TxState;                     // current state of Tx state machine
   BYTE_T b;

   ISR_PROLOG();
   pLn  = Line[1];
   Status = INB(LN01_IO) & pLn->WR15_; // read Ext status, enabled ESC's only
   if(Status & TXUNDERRUN_COND)        // a Tx underrun occurred
   {
      TxState = pLn->TxState;          // Tx state machine, current state
      if(TxState == STX_EOF_WAIT)      // valid end of Tx frame
      {
         // Disable Tx underrun interrups
         pLn->WR15_ &= ~TXUNDERRUN_IE;
         b = pLn->WR15_;
         ASM {
            push dx
            mov  dx,LN01_IO
            mov  al,WR15
            out  dx,al
            mov  al,b
            out  dx,al
            pop  dx
         }
         pLn->TxState = STX_EOF;       // state machine to Tx EOF processing
      }
      else                             // aborted Tx frame
      {
         DIS_DMA(LN1_DISTXDMA);        // disable Tx DMA channel
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN01_IO,WR15);
         OUTB(LN01_IO,pLn->WR15_);
         OUTB(LN01_IO,SEND_ABORT);     // send an abort
         Trace(0xc301,pLn->TxbqOut,TxState,pLn->TxbqRetryCnt);
         pLn->TxState = STX_RETRY;     // state machine to Tx retry state
      }
      OUTB(LN01_IO,RESET_TXUNDERRUN);  // reset Tx underrun latch
   }

   if(Status & ABORT_COND)             // Rx abort sequence occurred
   {
      // This frame will be discarded, re-initialize DMA so the next frame
      // will be placed in the same buffer this frame was going in.  Put
      // SCC back in Hunt mode so that no more data will be received until
      // a new frame starts.
      pBuf = &pLn->Rxb[pLn->RxbqIn];   // get ptr to buffer struct
      OUTB(DCH,LN1_RXDMACH);           // set up to access Rx DMA channel
      OUTW(DCA,pBuf->BufLo);           // Rx buffer absolute low word
      OUTB(DCA+2,pBuf->BufHi);         // Rx buffer absolute high word
      OUTW(DCC,RXBQ_BUF_SIZE - 1);     // xfer cnt, xfers 1 more than DCC
      OUTB(LN01_IO,WR3);               // put receiver in hunt mode
      OUTB(LN01_IO,pLn->WR3_ | ENTER_HUNT_MODE);

      SysMsg[0] = SYSCMD_RXERROR;      // message to SYS, cmd field
      SysMsg[1] = 1;                   //    line number field
      SysMsg[2] = MSG_RXERR_ABORT;     //    error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc401,pLn->RxbqIn,0,0);
   }

   // SCC: reset external status interrrupts and end of interrupt
   // PIC: end of interrupt
   ASM {
      mov  dx,LN01_IO
      mov  al,RESET_EXT
      out  dx,al
      mov  al,RESET_IUS
      out  dx,al
      mov  InIsr,0
      mov  dx,INTCTL
      mov  al,EOIVAL
      out  dx,al
   }
}
#endif // #if LN1_HDLC

#if LN0_BISYNC
void interrupt far Ln0EscIsr()
{
   LINE_T far *pLn;
   UBYTE_T Status;                     // SCC Tx, Rx, and external status
   WORD_T TxState;                     // current state of Tx state machine

   ISR_PROLOG();
   pLn  = Line[0];
   Status = INB(LN00_IO) & pLn->WR15_; // read Ext status, enabled ESC's only

   if(Status & TXUNDERRUN_COND)        // a Tx underrun occurred
   {
      TxState = pLn->TxState;          // Tx state machine, current state
      if(TxState == STX_EOF_WAIT)      // valid end of Tx frame
      {
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN00_IO,WR15);
         OUTB(LN00_IO,pLn->WR15_);
         pLn->TxState = STX_EOF;       // state machine to Tx EOF processing
      }
      else                             // aborted Tx frame
      {
         DIS_DMA(LN0_DISTXDMA);        // disable Tx DMA channel
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN00_IO,WR15);
         OUTB(LN00_IO,pLn->WR15_);
         Trace(0xc500,pLn->TxbqOut,TxState,pLn->TxbqRetryCnt);
         pLn->TxState = STX_RETRY;     // state machine to Tx retry state
      }
      OUTB(LN00_IO,RESET_TXUNDERRUN);  // reset Tx underrun latch
   }

   OUTB(LN00_IO,RESET_EXT);            // WR0, reset external status interrupts
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif // #if LN0_BISYNC

#if LN1_BISYNC
void interrupt far Ln1EscIsr()
{
   LINE_T far *pLn;
   UBYTE_T Status;                     // SCC Tx, Rx, and external status
   WORD_T TxState;                     // current state of Tx state machine

   ISR_PROLOG();
   pLn  = Line[1];
   Status = INB(LN01_IO) & pLn->WR15_; // read Ext status, enabled ESC's only
   if(Status & TXUNDERRUN_COND)        // a Tx underrun occurred
   {
      TxState = pLn->TxState;          // Tx state machine, current state
      if(TxState == STX_EOF_WAIT)      // valid end of Tx frame
      {
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN01_IO,WR15);
         OUTB(LN01_IO,pLn->WR15_);
         pLn->TxState = STX_EOF;       // state machine to Tx EOF processing
      }
      else                             // aborted Tx frame
      {
         DIS_DMA(LN1_DISTXDMA);        // disable Tx DMA channel
         pLn->WR15_ &= ~TXUNDERRUN_IE; // disable Tx underrun ints
         OUTB(LN01_IO,WR15);
         OUTB(LN01_IO,pLn->WR15_);
         Trace(0xc501,pLn->TxbqOut,TxState,pLn->TxbqRetryCnt);
         pLn->TxState = STX_RETRY;     // state machine to Tx retry state
      }
      OUTB(LN01_IO,RESET_TXUNDERRUN);  // reset Tx underrun latch
   }

   OUTB(LN01_IO,RESET_EXT);            // WR0, reset external status interrupts
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif // #if LN1_BISYNC


/*****************************************************************************
Function: LnXRcaIsr
Purpose:  Receive Character Available Interrupt Service Routines, one for
          each line.
Comments: HDLC uses DMA Rx with "Recevie Interrupt on Special Condition only"
          mode, so you would not expect it to need an RCA ISR.
          However, the "Recevie Interrupt on Special Condition only" mode can
          actually generate an RCA interrupt under certain conditions, see
          Am85C30 SCC 1992 Technical Manual, section 3.6.4.  To cover this
          case SRC ISR is vectored to if an RCA interrupt occurs.

          Async and Bisync Rx both use interrupt driven Rx and so need an
          RCA ISR.  Both take advantage of the SCC Rx FIFO by reading until
          the SCC Rx buffer is empty.  Doing so requires that the SRC status
          (RR1) be checked in the RCA ISR.  Since this means that SRC must be
          handled here, there is no seperate SRC ISR.  Instead, SRC interrupts
          are vectored to the RCA ISR.
*/
#if LN0_ASYNC
void interrupt far Ln0RcaIsr()
{
   LINE_T far *pLn;
   BYTE_T Ch;                          // character read from SCC
   WORD_T Out, In;                     // Rx buffer In & Out indices
   WORD_T NumFull;                     // number Rx queue locations filled
   UBYTE_T Status;                     // SRC status

   ISR_PROLOG();
   pLn  = Line[0];

   while(INB(LN00_IO) & RX_BUFFER)     // while Rx data available
   {
      // Process any errors
      OUTB(LN00_IO,RR1);               // get SRC status
      Status = INB(LN00_IO) & (PARITY_ERROR | OVERRUN_ERROR | FRAMING_ERROR);
      Ch = INB(LN00_IO+2);             // read character from SCC
      if(Status)                       // errors
      {
         SysMsg[0] = SYSCMD_RXERROR;   // msg to SYS, command field
         SysMsg[1] = 0;                //    line number field
         if(Status & OVERRUN_ERROR)    // Rx overrun in SCC
         {
            SysMsg[2] = MSG_RXERR_OVERRUN; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }
         if(Status & FRAMING_ERROR)    // framing error
         {
            SysMsg[2] = MSG_RXERR_FRAME; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }
         else if(Status & PARITY_ERROR) // parity error
         {
            SysMsg[2] = MSG_RXERR_PARITY; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }

         OUTB(LN00_IO,ERROR_RESET);    // WR0, error reset command
         SysIntPend = 1;               // queue interrupt to SYS for messages
      }

      // Process character
      Out = GetRxqOut(pLn);
      In = pLn->RxqIn;
      if((NumFull = In - Out) < 0)     // num queue locations full
      {
         NumFull += RXQ_SIZE;          // adjust for queue wrap
      }
      if(NumFull < RXQ_SIZE - 1)       // if Rx queue has empty space
      {
         *(pLn->RxqCtl + In) = Ch;     // add received character to queue
         pLn->RxqIn = (In + 1) % RXQ_SIZE; // bump In pointer
      }
      else                             // Rx queue overflow, char lost
      {
         Trace(pLn->LineNum | 0x3000,In,Out,pLn->RxqOut);
         SysMsg[0] = SYSCMD_RXERROR;   // msg to SYS, command field
         SysMsg[1] = 0;                //    line number field
         SysMsg[2] = MSG_RXERR_OVERFLOW; //  error flag field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
      }
   }

   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_ASYNC
void interrupt far Ln1RcaIsr()
{
   LINE_T far *pLn;
   BYTE_T Ch;                          // character read from SCC
   WORD_T Out, In;                     // Rx buffer In & Out indices
   WORD_T NumFull;                     // number Rx queue locations filled
   BYTE_T Status;                      // SRC status

   ISR_PROLOG();
   pLn  = Line[1];

   while(INB(LN01_IO) & RX_BUFFER)     // while Rx data available
   {
      // Process any errors
      OUTB(LN01_IO,RR1);               // get SRC status
      Status = INB(LN01_IO) & (PARITY_ERROR | OVERRUN_ERROR | FRAMING_ERROR);
      Ch = INB(LN01_IO+2);             // read character from SCC
      if(Status)                       // no errors
      {
         SysMsg[0] = SYSCMD_RXERROR;   // msg to SYS, command field
         SysMsg[1] = 1;                //    line number field
         if(Status & OVERRUN_ERROR)    // Rx overrun in SCC
         {
            SysMsg[2] = MSG_RXERR_OVERRUN; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }
         if(Status & FRAMING_ERROR)    // framing error
         {
            SysMsg[2] = MSG_RXERR_FRAME; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }
         else if(Status & PARITY_ERROR) // parity error
         {
            SysMsg[2] = MSG_RXERR_PARITY; // message to SYS, error flag field
            EnqSysMsg(SysMsg);         // send message to system
         }

         OUTB(LN01_IO,ERROR_RESET);    // WR0, error reset command
         SysIntPend = 1;               // queue interrupt to SYS for messages
      }

      // Process character
      Out = GetRxqOut(pLn);
      In = pLn->RxqIn;
      if((NumFull = In - Out) < 0)     // num queue locations full
      {
         NumFull += RXQ_SIZE;          // adjust for queue wrap
      }
      if(NumFull < RXQ_SIZE - 1)       // if Rx queue has empty space
      {
         *(pLn->RxqCtl + In) = Ch;     // add received character to queue
         pLn->RxqIn = (In + 1) % RXQ_SIZE; // bump In pointer
      }
      else                             // Rx queue overflow, char lost
      {
         SysMsg[0] = SYSCMD_RXERROR;   // msg to SYS, command field
         SysMsg[1] = 1;                //    line number field
         SysMsg[2] = MSG_RXERR_OVERFLOW; //  error flag field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
      }
   }

   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_HDLC
void interrupt far Ln0RcaIsr()
{
   ISR_PROLOG();
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_HDLC
void interrupt far Ln1RcaIsr()
{
   ISR_PROLOG();
   OUTB(LN01_IO,RESET_IUS);             // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_BISYNC
void interrupt far Ln0RcaIsr()
{
   LINE_T far *pLn;                    // ptr to line structure
   BYTE_T Ch;                          // character read from SCC
   WORD_T RxState;                     // current state of Rx state machine
   BYTE_T Status;                      // SRC status
   WORD_T Cnt;                         // buffer count
   WORD_T In;                          // buffer queue In index

   ISR_PROLOG();
   pLn  = Line[0];
   RxState = pLn->RxState;             // get current Rx state machine state
   switch(RxState)                     // state machine
   {
      case BRX_ACTIVE:                 // receiving active frame
         Ch = INB(LN00_IO+2);          // read character from SCC
         if(Ch != ETX)                 // data char for frame
         {
            Cnt = pLn->RxbqCurCnt;
            pLn->RxbqCurCnt = Cnt + 1;
            if(Cnt >= RXBQ_BUF_SIZE)   // overflow Rx buffer
            {
               SysMsg[2] = MSG_RXERR_OVERFLOW; // msg to SYS, error flag field
               goto HandleError;
            }
            pLn->RxbqCurBuf[Cnt] = Ch; // add char to buffer
         }
         else                          // end of frame
            pLn->RxState = BRX_CRC1_WAIT; // next state wait for CRC
         goto ExitIsr;
      case BRX_SOH_WAIT:               // waiting for SOH to start frame
         // Reset the CRC checker before reading 1st char so that it will be
         // included in the CRC computation.
         OUTB(LN00_IO,RESET_RXCRC);
         pLn->WR3_ |= RX_CRC_ENABLE;
         OUTB(LN00_IO,WR3);
         OUTB(LN00_IO,pLn->WR3_);
         if(INB(LN00_IO+2) == SOH)     // start of new frame
         {
            pLn->RxState = BRX_ACTIVE;
         }
         goto ExitIsr;
      case BRX_CRC1_WAIT:              // waiting for 1st CRC char
         INB(LN00_IO+2);               // discard char
         pLn->RxState = BRX_CRC2_WAIT;
         goto ExitIsr;
      case BRX_CRC2_WAIT:              // waiting for 2nd CRC char
         // Got CRC, but we must wait 2 more character times before the CRC
         // error status in RR1 becomes valid.  Since there must be at least
         // 2 SYNC chars between frames will use these to time the delay
         // (see Am85C30 SCC 1992 Technical Manual, section 4.10.1.5.4).
         INB(LN00_IO+2);               // discard char
         pLn->RxState = BRX_CRC3_WAIT;
         goto ExitIsr;
      case BRX_CRC3_WAIT:              // waiting for 1st post CRC char
         INB(LN00_IO+2);               // discard char
         pLn->RxState = BRX_CRC4_WAIT;
         goto ExitIsr;
      case BRX_CRC4_WAIT:              // waiting for 2nd post CRC char
         OUTB(LN00_IO,RR1);            // get SRC status
         Status = INB(LN00_IO);
         INB(LN00_IO+2);               // discard char
         if(Status & CRC_ERROR)        // bad CRC
         {
            SysMsg[2] = MSG_RXERR_CRC; // msg to SYS, error flag field
            Trace(0xc600,Ch,Status,pLn->RxbqIn);
            goto HandleError;
         }

         // If we get here the complete frame has been received with good CRC.
         // Update Rx buffer queue In index and count.  Need to protect
         // RxbCnt++ with dual port semaphore because the ++ operation is not
         // atomic and there is the possibliity that both system and controller
         // could be changing it simultaneously.
         //
         // The easiest way to get CRC checker ready for a new frame is to
         // enter hunt mode, but can't do so here because we already wasted
         // the two SYNC chars waiting for CRC status.  If we entered hunt mode
         // here we couldn't handle back to back frames with only one SYNC pair
         // between them.  Instead we will reset the CRC checker when the
         // next char arrives.
         In = pLn->RxbqIn;
         pLn->Rxb[In].Cnt = pLn->RxbqCurCnt;
         pLn->RxbqCurCnt = 0;
         In = (In + 1) % RXBQ_SIZE; // enq the buffer
         pLn->RxbqIn = In;
         pLn->RxbqCurBuf = pLn->Rxb[In].BufCtl;
         GET_DPSEM();
         pLn->RxbqCnt++;
         REL_DPSEM();
         SysIntPend = 1;               // queue an interrupt to the system
         pLn->WR3_ &= ~RX_CRC_ENABLE;  // disable SCC CRC checking
         if(pLn->RxbqCnt >= RXBQ_SIZE) // no buffers left
         {                             // must wait till buf becomes avail
            OUTB(LN00_IO,WR3);
            OUTB(LN00_IO,pLn->WR3_ & ~RX_ENABLE); // disable receiver
            pLn->RxState = BRX_BUF_WAIT; // Rx state machine - wait for buf
            Trace(0xc700,0,0,0);
         }
         else                          // more buffers available
         {
            OUTB(LN00_IO,WR3);
            OUTB(LN00_IO,pLn->WR3_);
            pLn->RxState = BRX_SOH_WAIT;  // ready for next frame
         }
         goto ExitIsr;
      case BRX_BUF_WAIT:
         INB(LN00_IO+2);               // discard char
         goto ExitIsr;
      default:                         // invalid state machine state
         Close(pLn);
         SysMsg[0] = SYSCMD_STATEERROR; // msg to SYS, command field
         SysMsg[1] = 0;                 //    line number field
         SysMsg[2] = 0;
         EnqSysMsg(SysMsg);             // send message to system
         SysIntPend = 1;                // queue an interrupt to the system
         goto ExitIsr;
   } // end switch(RxState)

HandleError:                           // error processing
   pLn->RxbqCurCnt = 0;                // discard frame
   Ch = INB(LN00_IO);                  // discard any data in SCC Rx FIFO
   Ch = INB(LN00_IO);
   Ch = INB(LN00_IO);
   OUTB(LN00_IO,WR3);                  // find start of next frame
   OUTB(LN00_IO,pLn->WR3_ | ENTER_HUNT_MODE);
   SysMsg[0] = SYSCMD_RXERROR;         // msg to SYS, command field
   SysMsg[1] = 0;                      //    line number field
   EnqSysMsg(SysMsg);                  // send message to system
   SysIntPend = 1;                     // queue an interrupt to the system
   pLn->RxState = BRX_SOH_WAIT;        // next - wait for new frame

ExitIsr:
   OUTB(LN00_IO,RESET_IUS);               // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif //#if LN0_BISYNC

#if LN1_BISYNC
void interrupt far Ln1RcaIsr()
{
   LINE_T far *pLn;                    // ptr to line structure
   BYTE_T Ch;                          // character read from SCC
   WORD_T RxState;                     // current state of Rx state machine
   BYTE_T Status;                      // SRC status
   WORD_T Cnt;                         // buffer count
   WORD_T In;                          // buffer queue In index

   ISR_PROLOG();
   pLn  = Line[1];
   RxState = pLn->RxState;             // get current Rx state machine state
   switch(RxState)                     // state machine
   {
      case BRX_ACTIVE:                 // receiving active frame
         Ch = INB(LN01_IO+2);          // read character from SCC
         if(Ch != ETX)                 // data char for frame
         {
            Cnt = pLn->RxbqCurCnt;
            pLn->RxbqCurCnt = Cnt + 1;
            if(Cnt >= RXBQ_BUF_SIZE)   // overflow Rx buffer
            {
               SysMsg[2] = MSG_RXERR_OVERFLOW; // msg to SYS, error flag field
               goto HandleError;
            }
            pLn->RxbqCurBuf[Cnt] = Ch; // add char to buffer
         }
         else                          // end of frame
            pLn->RxState = BRX_CRC1_WAIT; // next state wait for CRC
         goto ExitIsr;
      case BRX_SOH_WAIT:               // waiting for SOH to start frame
         // Reset the CRC checker before reading 1st char so that it will be
         // included in the CRC computation.
         OUTB(LN01_IO,RESET_RXCRC);
         pLn->WR3_ |= RX_CRC_ENABLE;
         OUTB(LN01_IO,WR3);
         OUTB(LN01_IO,pLn->WR3_);
         if(INB(LN01_IO+2) == SOH)     // start of new frame
         {
            pLn->RxState = BRX_ACTIVE;
         }
         goto ExitIsr;
      case BRX_CRC1_WAIT:              // waiting for 1st CRC char
         INB(LN01_IO+2);               // discard char
         pLn->RxState = BRX_CRC2_WAIT;
         goto ExitIsr;
      case BRX_CRC2_WAIT:              // waiting for 2nd CRC char
         // Got CRC, but we must wait 2 more character times before the CRC
         // error status in RR1 becomes valid.  Since there must be at least
         // 2 SYNC chars between frames will use these to time the delay
         // (see Am85C30 SCC 1992 Technical Manual, section 4.10.1.5.4).
         INB(LN01_IO+2);               // discard char
         pLn->RxState = BRX_CRC3_WAIT;
         goto ExitIsr;
      case BRX_CRC3_WAIT:              // waiting for 1st post CRC char
         INB(LN01_IO+2);               // discard char
         pLn->RxState = BRX_CRC4_WAIT;
         goto ExitIsr;
      case BRX_CRC4_WAIT:              // waiting for 2nd post CRC char
         OUTB(LN01_IO,RR1);            // get SRC status
         Status = INB(LN01_IO);
         INB(LN01_IO+2);               // discard char
         if(Status & CRC_ERROR)        // bad CRC
         {
            SysMsg[2] = MSG_RXERR_CRC; // msg to SYS, error flag field
            Trace(0xc601,Ch,Status,pLn->RxbqIn);
            goto HandleError;
         }

         // If we get here the complete frame has been received with good CRC.
         // Update Rx buffer queue In index and count.  Need to protect
         // RxbCnt++ with dual port semaphore because the ++ operation is not
         // atomic and there is the possibliity that both system and controller
         // could be changing it simultaneously.
         //
         // The easiest way to get CRC checker ready for a new frame is to
         // enter hunt mode, but can't do so here because we already wasted
         // the two SYNC chars waiting for CRC status.  If we entered hunt mode
         // here we couldn't handle back to back frames with only one SYNC pair
         // between them.  Instead we will reset the CRC checker when the
         // next char arrives.
         In = pLn->RxbqIn;
         pLn->Rxb[In].Cnt = pLn->RxbqCurCnt;
         pLn->RxbqCurCnt = 0;
         In = (In + 1) % RXBQ_SIZE; // enq the buffer
         pLn->RxbqIn = In;
         pLn->RxbqCurBuf = pLn->Rxb[In].BufCtl;
         GET_DPSEM();
         pLn->RxbqCnt++;
         REL_DPSEM();
         SysIntPend = 1;               // queue an interrupt to the system
         pLn->WR3_ &= ~RX_CRC_ENABLE;  // disable SCC CRC checking
         if(pLn->RxbqCnt >= RXBQ_SIZE) // no buffers left
         {                             // must wait till buf becomes avail
            OUTB(LN01_IO,WR3);
            OUTB(LN01_IO,pLn->WR3_ & ~RX_ENABLE); // disable receiver
            pLn->RxState = BRX_BUF_WAIT; // Rx state machine - wait for buf
            Trace(0xc701,0,0,0);
         }
         else                          // more buffers available
         {
            OUTB(LN01_IO,WR3);
            OUTB(LN01_IO,pLn->WR3_);
            pLn->RxState = BRX_SOH_WAIT;  // ready for next frame
         }
         goto ExitIsr;
      case BRX_BUF_WAIT:
         INB(LN01_IO+2);               // discard char
         goto ExitIsr;
      default:                         // invalid state machine state
         Close(pLn);
         SysMsg[0] = SYSCMD_STATEERROR; // msg to SYS, command field
         SysMsg[1] = 1;                 //    line number field
         SysMsg[2] = 0;                 //    line number field
         EnqSysMsg(SysMsg);             // send message to system
         SysIntPend = 1;                // queue an interrupt to the system
         goto ExitIsr;
   } // end switch(RxState)

HandleError:                           // error processing
   pLn->RxbqCurCnt = 0;                // discard frame
   Ch = INB(LN01_IO);                  // discard any data in SCC Rx FIFO
   Ch = INB(LN01_IO);
   Ch = INB(LN01_IO);
   OUTB(LN01_IO,WR3);                  // find start of next frame
   OUTB(LN01_IO,pLn->WR3_ | ENTER_HUNT_MODE);
   SysMsg[0] = SYSCMD_RXERROR;         // msg to SYS, command field
   SysMsg[1] = 1;                      //    line number field
   EnqSysMsg(SysMsg);                  // send message to system
   SysIntPend = 1;                     // queue an interrupt to the system
   pLn->RxState = BRX_SOH_WAIT;        // next - wait for new frame

ExitIsr:
   OUTB(LN01_IO,RESET_IUS);               // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

/*****************************************************************************
Function: LnXSrcIsr
Purpose:  Special Receive Condition Interrupt Service Routines, one for each
          line.
Comments: For Async and Bisync SRC interrupts are handled in the RCA ISR, so
          SRC interrupts are vectored there.
*/
#if LN0_ASYNC
void interrupt far Ln0SrcIsr()
{
   ISR_PROLOG();
   OUTB(LN00_IO,ERROR_RESET);          // WR0, error reset command
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_ASYNC
void interrupt far Ln1SrcIsr()
{
   ISR_PROLOG();
   OUTB(LN01_IO,ERROR_RESET);          // WR0, error reset command
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN0_HDLC
void interrupt far Ln0SrcIsr()
{
   LINE_T far *pLn;
   UBYTE_T Status;                     // SRC status
   WORD_T In;                          // buffer queue In index
   BUFFER_T far *pBuf;                 // ptr to buffer structure
   WORD_T RestartDma;                  // set if DMA to be restarted
   WORD_T w;
   BYTE_T b;

   ISR_PROLOG();
   pLn  = Line[0];

   // Disable Rx DMA channel, read SRC status
   ASM {
      push dx
      mov  dx,DDC
      mov  al,DDMA
      out  dx,al
      mov  dx,DMK
      in   al,dx
      or   al,LN0_DISRXDMA
      out  dx,al
      mov  dx,DDC
      mov  al,0
      out  dx,al

      mov  dx,LN00_IO
      mov  al,RR1
      out  dx,al
      in   al,dx
      mov  Status,al
      pop  dx
   }

   In = pLn->RxbqIn;                   // get current Rx buffer index
   RestartDma = 1;                     // assume DMA is to be restarted

   if(Status & OVERRUN_ERROR)          // Rx overrun error
   {
      SysMsg[0] = SYSCMD_RXERROR;      // message to SYS, cmd field
      SysMsg[1] = 0;                   //    line number field
      SysMsg[2] = MSG_RXERR_OVERRUN;   //    error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc800,0,0,0);
   }
   else if(Status & END_FRAME)         // end of Rx frame
   {
      if(Status & CRC_ERROR)           // CRC error occurred in frame
      {
         SysMsg[0] = SYSCMD_RXERROR;   // message to SYS, cmd field
         SysMsg[1] = 0;                //    line number field
         SysMsg[2] = MSG_RXERR_CRC;    //    error flag field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
         Trace(0xc900,0,0,0);
      }
      else                             // good frame
      {
#if !TEST_HDLC                         // normal HDLC processing
         // Enqueue this buffer
         In = pLn->RxbqIn;
            // get number bytes in read from DMA channel, -3 eliminates CRC
         ASM {
            push dx
            mov  dx,DCH
            mov  al,LN0_RXDMACH
            out  dx,al
            mov  dx,DCC
            in   ax,dx
            mov  w,ax
            pop  dx
         }
         pLn->Rxb[In].Cnt = RXBQ_BUF_SIZE - 3 - w;

         In = (In + 1) % RXBQ_SIZE;    // bump In index
         pLn->RxbqIn = In;
         GET_DPSEM();                  // ++ operation not atomic, must protect
         pLn->RxbqCnt++;               // bump buf count
         REL_DPSEM();
         SysIntPend = 1;               // queue an interrupt to the system
         if(pLn->RxbqCnt >= RXBQ_SIZE) // no buffers left
         {                             // must wait till buf becomes avail
            // Disable receiver
            b = pLn->WR3_ & ~RX_ENABLE;
            ASM {
               push dx
               mov  dx,LN00_IO
               mov  al,WR3
               out  dx,al
               mov  al,b
               out  dx,al
               pop  dx
            }

            pLn->RxState = HRX_BUF_WAIT; // Rx state machine - wait for buf
            RestartDma = 0;            // don't restart the DMA
            Trace(0xca00,0,0,0);
         }
#else
         OUTB(DCH,LN0_RXDMACH);        // set up to access Rx dma channel
         In = RXBQ_BUF_SIZE - 3 - INW(DCC); // num bytes read, -3 eliminate CRC
         if(In != T_TXLEN)             // bad num bytes read
         {
            Trace(0xcb00,In,pLn->RxbqIn,0);
            In = (In + 1) % RXBQ_SIZE;
            pLn->RxbqIn = In;
//          Close(Line[0]);            // halt test
//          Close(Line[1]);
         }
         else
         {
            In = pLn->RxbqIn;
            pBuf = &pLn->Rxb[In];      // get ptr to buffer struct
            if(pBuf->BufCtl[1] != 0x30) // bad line number
            {
               Trace(0xcc00,0,0,0);
               In = (In + 1) % RXBQ_SIZE;
               pLn->RxbqIn = In;
//             Close(Line[0]);         // halt test
//             Close(Line[1]);
            }
            else if(pBuf->BufCtl[2] != In+0x30) // bad sequence number
            {
               Trace(0xcd00,In+0x30,pBuf->BufCtl[2],0);
               In = pLn->TxbqOut;      // resync Rx frame num to Tx frame num
               pLn->RxbqIn = In;
//             Close(Line[0]);         // halt test
//             Close(Line[1]);
            }
            else                       // good Rx frame
            {
               // bump In index
               In = (In + 1) % RXBQ_SIZE;
               pLn->RxbqIn = In;
            }
         }
#endif
      }
   }

   if(RestartDma)
   {
      // Configure DMA channel for next Rx frame
      pBuf = &pLn->Rxb[In];            // get ptr to buffer struct
      w = pBuf->BufLo;                      // Rx buffer absolute low word
      b = pBuf->BufHi;                      // Rx buffer absolute high byte
      ASM {
         push dx
         mov  dx,DCH
         mov  al,LN0_RXDMACH
         out  dx,al
         mov  dx,DCA
         mov  ax,w
         out  dx,ax
         inc  dx
         inc  dx
         mov  al,b
         out  dx,al
         mov  dx,DCC
         mov  ax,RXBQ_BUF_SIZE - 1
         out  dx,ax

         mov  dx,DDC
         mov  al,DDMA
         out  dx,al
         mov  dx,DMK
         in   al,dx
         and  al,LN0_ENRXDMA
         out  dx,al
         mov  dx,DDC
         mov  al,0
         out  dx,al
         pop  dx
      }
   }

   // SCC: error reset and end of interrupt
   // PIC: end of interrupt
   ASM {
      mov  dx,LN00_IO
      mov  al,ERROR_RESET
      out  dx,al
      mov  al,RESET_IUS
      out  dx,al
      mov  InIsr,0
      mov  dx,INTCTL
      mov  al,EOIVAL
      out  dx,al
   }
}
#endif

#if LN1_HDLC
void interrupt far Ln1SrcIsr()
{
   LINE_T far *pLn;
   UBYTE_T Status;                     // SRC status
   WORD_T In;                          // buffer queue In index
   BUFFER_T far *pBuf;                 // ptr to buffer structure
   WORD_T RestartDma;                  // set if DMA to be restarted
   WORD_T w;
   BYTE_T b;

   ISR_PROLOG();
   pLn  = Line[1];

   // Disable Rx DMA channel, read SRC status
   ASM {
      push dx
      mov  dx,DDC
      mov  al,DDMA
      out  dx,al
      mov  dx,DMK
      in   al,dx
      or   al,LN1_DISRXDMA
      out  dx,al
      mov  dx,DDC
      mov  al,0
      out  dx,al

      mov  dx,LN01_IO
      mov  al,RR1
      out  dx,al
      in   al,dx
      mov  Status,al
      pop  dx
   }

   In = pLn->RxbqIn;                   // get current Rx buffer index
   RestartDma = 1;                     // assume DMA is to be restarted

   if(Status & OVERRUN_ERROR)          // Rx overrun error
   {
      SysMsg[0] = SYSCMD_RXERROR;      // message to SYS, cmd field
      SysMsg[1] = 1;                   //    line number field
      SysMsg[2] = MSG_RXERR_OVERRUN;   //    error flag field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      Trace(0xc801,0,0,0);
   }
   else if(Status & END_FRAME)         // end of Rx frame
   {
      if(Status & CRC_ERROR)           // CRC error occurred in frame
      {
         SysMsg[0] = SYSCMD_RXERROR;   // message to SYS, cmd field
         SysMsg[1] = 1;                //    line number field
         SysMsg[2] = MSG_RXERR_CRC;    //    error flag field
         EnqSysMsg(SysMsg);            // send message to system
         SysIntPend = 1;               // queue an interrupt to the system
         Trace(0xc901,0,0,0);
      }
      else                             // good frame
      {
#if !TEST_HDLC                         // normal HDLC processing
         // Enqueue this buffer
         In = pLn->RxbqIn;
            // get number bytes in read from DMA channel, -3 eliminates CRC
         ASM {
            push dx
            mov  dx,DCH
            mov  al,LN1_RXDMACH
            out  dx,al
            mov  dx,DCC
            in   ax,dx
            mov  w,ax
            pop  dx
         }
         pLn->Rxb[In].Cnt = RXBQ_BUF_SIZE - 3 - w;

         In = (In + 1) % RXBQ_SIZE;    // bump In index
         pLn->RxbqIn = In;
         GET_DPSEM();                  // ++ operation not atomic, must protect
         pLn->RxbqCnt++;               // bump buf count
         REL_DPSEM();
         SysIntPend = 1;               // queue an interrupt to the system
         if(pLn->RxbqCnt >= RXBQ_SIZE) // no buffers left
         {                             // must wait till buf becomes avail
            // Disable receiver
            b = pLn->WR3_ & ~RX_ENABLE;
            ASM {
               push dx
               mov  dx,LN01_IO
               mov  al,WR3
               out  dx,al
               mov  al,b
               out  dx,al
               pop  dx
            }
            pLn->RxState = HRX_BUF_WAIT; // Rx state machine - wait for buf
            RestartDma = 0;            // don't restart the DMA
            Trace(0xca01,0,0,0);
         }
#else
         OUTB(DCH,LN1_RXDMACH);        // set up to access Rx dma channel
         In = RXBQ_BUF_SIZE - 3 - INW(DCC); // num bytes read, -3 eliminate CRC
         if(In != T_TXLEN)             // bad num bytes read
         {
            Trace(0xcb01,In,pLn->RxbqIn,0);
            In = (In + 1) % RXBQ_SIZE;
            pLn->RxbqIn = In;
//          Close(Line[0]);            // halt test
//          Close(Line[1]);
         }
         else
         {
            In = pLn->RxbqIn;
            pBuf = &pLn->Rxb[In];      // get ptr to buffer struct
            if(pBuf->BufCtl[1] != 0x31) // bad line number
            {
               Trace(0xcc01,0,0,0);
               In = (In + 1) % RXBQ_SIZE;
               pLn->RxbqIn = In;
//             Close(Line[0]);         // halt test
//             Close(Line[1]);
            }
            else if(pBuf->BufCtl[2] != In+0x30) // bad sequence number
            {
               Trace(0xcd01,In+0x30,pBuf->BufCtl[2],0);
               In = pLn->TxbqOut;      // resync Rx frame num to Tx frame num
               pLn->RxbqIn = In;
//             Close(Line[0]);         // halt test
//             Close(Line[1]);
            }
            else                       // good Rx frame
            {
               // bump In index
               In = (In + 1) % RXBQ_SIZE;
               pLn->RxbqIn = In;
            }
         }
#endif
      }
   }

   if(RestartDma)
   {
      // Configure DMA channel for next Rx frame
      pBuf = &pLn->Rxb[In];            // get ptr to buffer struct
      w = pBuf->BufLo;                      // Rx buffer absolute low word
      b = pBuf->BufHi;                      // Rx buffer absolute high byte
      ASM {
         push dx
         mov  dx,DCH
         mov  al,LN1_RXDMACH
         out  dx,al
         mov  dx,DCA
         mov  ax,w
         out  dx,ax
         inc  dx
         inc  dx
         mov  al,b
         out  dx,al
         mov  dx,DCC
         mov  ax,RXBQ_BUF_SIZE - 1
         out  dx,ax

         mov  dx,DDC
         mov  al,DDMA
         out  dx,al
         mov  dx,DMK
         in   al,dx
         and  al,LN1_ENRXDMA
         out  dx,al
         mov  dx,DDC
         mov  al,0
         out  dx,al
         pop  dx
      }
   }

   // SCC: error reset and end of interrupt
   // PIC: end of interrupt
   ASM {
      mov  dx,LN01_IO
      mov  al,ERROR_RESET
      out  dx,al
      mov  al,RESET_IUS
      out  dx,al
      mov  InIsr,0
      mov  dx,INTCTL
      mov  al,EOIVAL
      out  dx,al
   }
}
#endif

#if LN0_BISYNC
void interrupt far Ln0SrcIsr()
{
   LINE_T far *pLn;                    // ptr to line structure
   BYTE_T Status;                      // SRC status

   ISR_PROLOG();
   OUTB(LN00_IO,RR1);                  // get SRC status
   Status = INB(LN00_IO);
   if(Status & OVERRUN_ERROR)          // Rx overrun
   {
      pLn  = Line[0];
      SysMsg[2] = MSG_RXERR_OVERRUN;   // message to SYS, error flag field
      Trace(0xce00,pLn->RxbqCurCnt,0,0);
      pLn->RxbqCurCnt = 0;             // discard frame
      INB(LN00_IO);                    // discard any data in SCC Rx FIFO
      INB(LN00_IO);
      INB(LN00_IO);
      OUTB(LN00_IO,WR3);               // find start of next frame
      OUTB(LN00_IO,pLn->WR3_ | ENTER_HUNT_MODE);
      SysMsg[0] = SYSCMD_RXERROR;      // msg to SYS, command field
      SysMsg[1] = 0;                   //    line number field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      pLn->RxState = BRX_SOH_WAIT;     // next - wait for new frame
   }
   OUTB(LN00_IO,ERROR_RESET);          // WR0, error reset command
   OUTB(LN00_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

#if LN1_BISYNC
void interrupt far Ln1SrcIsr()
{
   LINE_T far *pLn;                    // ptr to line structure
   BYTE_T Status;                      // SRC status

   ISR_PROLOG();
   OUTB(LN01_IO,RR1);                  // get SRC status
   Status = INB(LN01_IO);
   if(Status & OVERRUN_ERROR)          // Rx overrun
   {
      pLn  = Line[1];
      SysMsg[2] = MSG_RXERR_OVERRUN;   // message to SYS, error flag field
      Trace(0xce01,pLn->RxbqCurCnt,0,0);
      pLn->RxbqCurCnt = 0;             // discard frame
      INB(LN01_IO);                    // discard any data in SCC Rx FIFO
      INB(LN01_IO);
      INB(LN01_IO);
      OUTB(LN01_IO,WR3);               // find start of next frame
      OUTB(LN01_IO,pLn->WR3_ | ENTER_HUNT_MODE);
      SysMsg[0] = SYSCMD_RXERROR;      // msg to SYS, command field
      SysMsg[1] = 1;                   //    line number field
      EnqSysMsg(SysMsg);               // send message to system
      SysIntPend = 1;                  // queue an interrupt to the system
      pLn->RxState = BRX_SOH_WAIT;     // next - wait for new frame
   }
   OUTB(LN01_IO,ERROR_RESET);          // WR0, error reset command
   OUTB(LN01_IO,RESET_IUS);            // WR0, end of interrupt to SCC
   ISR_EPILOG();
}
#endif

/***************************************************************************
Function: GetCtlTime
Purpose:  Get the controller time.
Call:     GetCtlTime()
Return:   UDWORD_T: The controller time.
Comments: Controller time is the number of timer 1 ticks that have occurred
          since the control program was started, assuming the tick count
          has not rolled over.

          While getting the controller time (heartbeat) don't allow Timer1Isr()
          to go off.  This because Timer1Isr() increments the heartbeat value,
          which is a double word.  If this interrupt occurs between the
          word reads of heartbeat in GetCtlTime(), the value returned could
          be corrupted.
*/
UDWORD_T GetCtlTime(void)
{
   UDWORD_T Time;

   DIS_INT();               // don't let Timer1Isr() interrupt double word xfer
   Time = pFirm->heartbeat;
   EN_INT();
   return(Time);
}

/*****************************************************************************
Function:   DecrementSfg
Purpose:    Decrement an SFG and return it's state after decrementing.
Call:       DecrementSfg(Group)
            WORD_T Group;  Status flag group number, identifies which SFG
                           to operate on.
Return:     UWORD_T: The state of the SFG after decrementing.
Comments:   This function is useful if an SFG is to be used to count things.
            In this case the SFG would be thought of as a UWORD_T value rather
            than a group of 16 flags.
            See INIT_SFG.
Warnings:   This function gets and releases dual port access.
*/
UWORD_T DecrementSfg(WORD_T Group)
{
   UWORD_T State;                      // the state of the SFG

   DIS_INT();
   GET_DPSEM();
   State = --pSfg[Group];
   REL_DPSEM();
   EN_INT();
   return(State);
}

/*****************************************************************************
Function:   GetSfg
Purpose:    Get a masked copy of a Status Flag Group.
Call:       GetSFG(Group,Mask)
            WORD_T Group; Status flag group number, identifies which SFG
                          to operate on.
            UWORD_T Mask; Only the flags set in Mask have their states
                          returned, all flags not set in Mask are cleared
                          in the return value.   If the current state of all
                          flags is desired set Mask to ffffH.
Return:     UWORD_T: The masked copy of the SFG.
Comments:   See INIT_SFG.
Warnings:   This function gets and releases dual port access.
*/
UWORD_T GetSfg(WORD_T Group,UWORD_T Mask)
{
   UWORD_T State;                   // the state of the SFG

   DIS_INT();
   GET_DPSEM();
   State = pSfg[Group];
   REL_DPSEM();
   EN_INT();
   return(State & Mask);
}

/*****************************************************************************
Function:   IncrementSfg
Purpose:    Increment an SFG and return it's state after incrementing.
Call:       IncrementSfg(Group)
            WORD_T Group;  Status flag group number, identifies which SFG
                           to operate on.
Return:     UWORD_T: The state of the SFG after incrementing.
Comments:   This function is useful if an SFG is to be used to count things.
            In this case the SFG would be thought of as a UWORD_T value rather
            than a group of 16 flags.
            See INIT_SFG.
Warnings:   This function gets and releases dual port access.
*/
UWORD_T IncrementSfg(WORD_T Group)
{
   UWORD_T State;                      // the state of the SFG

   DIS_INT();
   GET_DPSEM();
   State = ++pSfg[Group];
   REL_DPSEM();
   EN_INT();
   return(State);
}

/*****************************************************************************
Function:   TestSetSfg
Purpose:    Test and set the state of a Status Flag Group.
Call:       TestSetSfg(Group,Mask)
            WORD_T Group;  Status flag group number, identifies which SFG
                           to operate on.
            UWORD_T Mask;  Or mask.  All flags set to 1 in Mask will be
                           set in the flag group, other flags will be
                           unchanged.
Return:     UWORD_T: The state of the SFG before the set operation was
                     performed.
Comments:   This function first saves the state of the SFG, then performs the
            set operation, then returns the saved state.
            See INIT_SFG.
Warnings:   This function gets and releases dual port access.
*/
UWORD_T TestSetSfg(WORD_T Group,UWORD_T Mask)
{
   UWORD_T State;                      // state of flag group

   DIS_INT();
   GET_DPSEM();
   State = pSfg[Group];
   pSfg[Group] |= Mask;
   REL_DPSEM();
   EN_INT();
   return(State);
}

/*****************************************************************************
Function: Trace
Purpose:  Put a four word trace entry into the trace buffer.
Call:     Trace(w1,w2,w3,w4)
          WORD_T w1; 1st word to put in trace buffer.
          WORD_T w2; 2nd word to put in trace buffer.
          WORD_T w3; 3rd word to put in trace buffer.
          WORD_T w4; 4th word to put in trace buffer.
Return:   void
Comments: The parameters are copied directly into the trace buffer at
          the index contained in pTrIx.
*/
void Trace(WORD_T w1,WORD_T w2,WORD_T w3,WORD_T w4)
{
   register WORD_T i;

   DIS_INT();
   i = *pTrIx;
   pTrBuf[i] = w1;
   pTrBuf[i+1] = w2;
   pTrBuf[i+2] = w3;
   pTrBuf[i+3] = w4;
   i = (i+4) % (TRBUF_SIZE/2);
   *pTrIx = i;
   EN_INT();
}

/*****************************************************************************
Function: ExecTestModes
Purpose:  Execute special test modes.
Call:     ExecTestModes()
Return:   void
Comments: This executes those tests which are enabled by setting #define
          TEST_XXXX in CPC.H.
*/
void ExecTestModes(void)
{
   UDWORD_T WaitTime;

   // Give SYS loader time to finish so we are absolutely sure there is no
   // interference from it.
   WaitTime = GetCtlTime() + 300;
   while(GetCtlTime() < WaitTime){}

// This for testing SFGs only.  If TEST_SFG1 or TEST_SFG2 are set normal
// sync or async communications won't work because the control program dead
// ends in function TestSfgX().
#if TEST_SFG1
   TestSfg1();
#endif
#if TEST_SFG2
   TestSfg2();
#endif

// This causes line 0 to continuously transmit asynchronously without
// interaction from the system.
#if TEST_ASYNC
   SetTestModeAsync();
#endif

// This causes line 0 and 1 to continuously transmit and receive HDLC without
// interaction from the system.
#if TEST_HDLC
   SetTestModeHdlc();
#endif
}

#if TEST_ASYNC
/*****************************************************************************
Function: SetTestModeAsync
Purpose:  Open line 0 for special async mode test.
Call:     SetTestModeAsync()
Return:   void
Comments: This will cause line 0 to immediately start transmitting a single
          asynchronous character (ASCII 0) continuously at 9600 baud.  No
          interaction with the system is required.
*/
void SetTestModeAsync(void)
{
   LINE_T far *pLn;                    // ptr to line structure
   static BYTE_T MsgBuf[MSG_LEN];

   pLn = Line[0];
   MsgBuf[0] = CTLCMD_OPENASYNC;
   MsgBuf[1] = 0;
   MsgBuf[2] = MSG_BAUD_9600;
   MsgBuf[3] = MSG_PAR_NONE;
   MsgBuf[4] = MSG_DATABIT_8;
   MsgBuf[5] = MSG_STOPBIT_1;
   MsgBuf[6] = MSG_MDM_DTR | MSG_MDM_RTS;
   cmdOpenAsync(MsgBuf);
   pLn->TxqIn = 1;                     // put char to xmit in Tx buffer
   pLn->TxqOut = 0;
   *pLn->TxqCtl = 0x30;
}
#endif // end #if TEST_ASYNC

#if TEST_HDLC
/*****************************************************************************
Function: SetTestModeHdlc
Purpose:  Open line 0 and 1 for special HDLC mode test.
Call:     SetTestModeAsync()
Return:   void
Comments: This will cause all lines to immediately start transmitting and
          receiving T_TXLEN byte frames.  The TxbqCnt and RxbqCnt fields
          are fixed so that it always appears that there are Tx and Rx
          buffers available.  The same buffers will be transmitted
          repeatadly, and the Rx buffers will be filled repeatadly.
*/
void SetTestModeHdlc(void)
{
   WORD_T Ln;
   WORD_T b,i;

   BUFFER_T far *pBuf;
   LINE_T far *pLn;
   static BYTE_T MsgBuf[MSG_LEN] = {CTLCMD_OPENHDLC,0,MSG_MDM_RTS | MSG_MDM_DTR,
      MSG_CLK_2,0,0,0,0,0,0,0,0,0,0,0,0};
   static BYTE_T TxStr[50] =
      "[XY]This is CTL side stand alone test of HDLC mode";

   MsgBuf[4] = (BYTE_T)((long)T_DATARATE);
   MsgBuf[5] = (BYTE_T)((long)T_DATARATE >> 8);
   MsgBuf[6] = (BYTE_T)((long)T_DATARATE >> 16);
   MsgBuf[7] = (BYTE_T)((long)T_DATARATE >> 24);

   for(Ln = 0;Ln < NUMLINES;Ln++)
   {
      pLn = Line[Ln];
      MsgBuf[1] = Ln;
      cmdOpenHdlc(MsgBuf);

      for(b = 0;b < TXBQ_SIZE;b++)     // for each Txbq buffer
      {
         pBuf = (BUFFER_T far *)&pLn->Txb[b];
         for(i = 0;i < T_TXLEN;i++)    // put some data in the Txbq buffer
         {
            pBuf->BufCtl[i] = TxStr[i % 50];
         }
         pBuf->BufCtl[i] = b + 0x40;   // this char should never transmit
         pBuf->BufCtl[1] = Ln + 0x30;  // put the line number in the buffer
         pBuf->BufCtl[2] = b + 0x30;   // put a sequence number in the buffer
         pBuf->Cnt = T_TXLEN;          // set Tx byte count
      }
      pLn->TxbqCnt = 8;                // make the line think it has Tx data
   }
}
#endif // end #if TEST_HDLC

#if TEST_SFG1
/*****************************************************************************
Function: TestSfg1
Purpose:  Test SFGs.
Call:     TestSfg1()
Return:   void - never returns
Comments: This is run in conjunction with Test3 on the system.

          SFG 0 and SFG 1 are used as a semaphores.  The system signals ready
          to the controller on SFG 0, and the controller signals ready to the
          system on SFG 1.
*/
void TestSfg1(void)
{
   BYTE_T i;                           // SFG group number
   UWORD_T Val;                        // value of SFG
   UWORD_T j;                          // ripple counter

   for(i = 0;i < SFG_SIZE;i++)         // startup state of all SFGs
      INIT_SFG(0,0);

   while(1)                            // test runs forever
   {
      // Test 1, controller initialize - system check
      while(GetSfg(0,0xffff) != 1);    // wait for system ready
      CLR_SFG(0,1);                    // clear system ready
      for(i = 2;i < SFG_SIZE;i++)      // initialize each SFG
      {
         INIT_SFG(i,0xaa00 + i);
      }
      SET_SFG(1,1);                    // signal controller ready

      // Test 2, controller clear - system check
      while(GetSfg(0,0xffff) != 1);    // wait for system ready
      CLR_SFG(0,1);                    // clear system ready
      for(i = 2;i < SFG_SIZE;i++)      // clear each SFG
      {
         CLR_SFG(i,0xffff);
      }
      SET_SFG(1,1);                    // signal controller ready

      // Test 3, controller progressively set flag bits - system check
      j = 1;                           // progressively set & check flag bits
      do
      {
         while(GetSfg(0,0xffff) != 1); // wait for system ready
         CLR_SFG(0,1);                 // clear system ready
         for(i = 2;i < SFG_SIZE;i++)
         {
            SET_SFG(i,j);              // set flag
         }
         j = j << 1;
         SET_SFG(1,1);                 // signal controller ready
      } while(j > 0);

      // Test 4, controller progressively clear flag bits - system check
      j = 0xfffe;
      do
      {
         while(GetSfg(0,0xffff) != 1); // wait for system ready
         CLR_SFG(0,1);                 // clear system ready
         for(i = 2;i < SFG_SIZE;i++)
         {
            CLR_SFG(i,~j);
         }
         j = (j << 1) + 1;
         SET_SFG(1,1);                 // signal controller ready
      } while(j < 0xffff);

      // Test 5, system initialize - controller check
      while(GetSfg(0,0xffff) != 1);    // wait for system ready
      CLR_SFG(0,1);                    // clear system ready
      for(i = 2;i < SFG_SIZE;i++)      // initialize each SFG
      {
         Val = 0xaa00 + i;
         if(GetSfg(i,0xffff) != Val)
            ProcError(1);
      }
      SET_SFG(1,1);                    // signal controller ready

      // Test 6, system clear - controller check
      while(GetSfg(0,0xffff) != 1);    // wait for system ready
      CLR_SFG(0,1);                    // clear system ready
      for(i = 2;i < SFG_SIZE;i++)      // clear each SFG
      {
         if(GetSfg(i,0xffff) != 0)
            ProcError(2);
      }
      SET_SFG(1,1);                    // signal controller ready

      // Test 7, system progressively set flag bits - controller check
      j = 1;                        // progressively set & check flag bits
      Val = 0;
      do
      {
         while(GetSfg(0,0xffff) != 1); // wait for system ready
         CLR_SFG(0,1);                 // clear system ready
         for(i = 2;i < SFG_SIZE;i++)
         {
            Val |= j;
            if(GetSfg(i,0xffff) != Val)
               ProcError(3);
         }
         j = j << 1;
         SET_SFG(1,1);                 // signal controller ready
      } while(j > 0);

      // Test 8, system progressively clear flag bits - controller check
      j = 0xfffe;
      do
      {
         while(GetSfg(0,0xffff) != 1); // wait for system ready
         CLR_SFG(0,1);                 // clear system ready
         for(i = 2;i < SFG_SIZE;i++)
         {
            Val &= j;
            if(GetSfg(i,0xffff) != Val)
               ProcError(4);
         }
         j = (j << 1) + 1;
         SET_SFG(1,1);                 // signal controller ready
      } while(j < 0xffff);
   }
}
#endif // #if TEST_SFG1

#if TEST_SFG2
/*****************************************************************************
Function: TestSfg2
Purpose:  Test TestSetSfg(), IncrementSfg(), DecrementSfg().
Call:     TestSfg2()
Return:   void - never returns
Comments: This is run in conjunction with Test4 on the system.

          SFG0 is used as a semaphore controlling access to a shared
          resource.  In this case the shared resource is SFG1.  Both the
          controller and the system compete for access to SFG1. Once access is
          granted SFG1 is incremented and checked by the winner.  SFG2 and
          3 are used to provide a visual indication of how often each
          side gets the resource, SFG2 is incremented each time the controller
          wins, and SFG3 each time the system wins.  SFG4 is decremented each
          time through the test loop just to exercise DecrementSfg().
*/
void TestSfg2(void)
{
   UWORD_T Val;                        // value of shared resource (SFG1)
   WORD_T i,j;

   for(i = 0;i < 4;i++)                // startup state of all SFGs
      INIT_SFG(i,0);

   while(1)                            // test runs forever
   {
      DecrementSfg(4);                 // exercise DecrementSfg()
      while(TestSetSfg(0,1));          // get semaphore
      IncrementSfg(2);                 // bump controller win count
      Val = 0;
      INIT_SFG(1,Val);                 // operate on shared resource
      while(Val < 0x40)                // operate on shared resource
      {
         IncrementSfg(1);
         Val++;
         for(j = 0;j < 5;j++)          // waste some time to allow system to
         {                             //    change it if semaphore not working
            i = j;
            i = i;
         }
         if(GetSfg(1,0xffff) != Val)   // shared resource collision
            ProcError(1);
      }
      CLR_SFG(0,1);                    // release semaphore
   }
}
#endif // #if TEST_SFG2

#if TEST_SFG1 || TEST_SFG2
/*****************************************************************************
Function: ProcError
Purpose:  Process an SFG test error.  Sets SFG 15 to 0xeeXX, where XX is an
          error code, then dead ends.
Call:     ProcErr(ErrorCode)
          UBYTE_T ErrorCode; Error code to place in SFG 15.
Return:   Void. Never returns.
*/
void ProcError(UBYTE_T ErrorCode)
{
   INIT_SFG(15,0xee00 + ErrorCode);
   while(1);                           // dead end
}
#endif

#ifdef BORLANDC
#pragma option -Od
#endif
#ifdef MSC
#pragma optimize("",off)
#endif
/***************************************************************************
Function: GetTxqIn
Purpose:  Get the Tx queue In index.
Call:     GetTxqIn(pLn)
          LINE_T far *pLn; Ptr to line structure
Return:   WORD_T: The Tx queue In index
Comments: This function repeatadly reads the In index until the values from
          subsequent reads match.  This is done to prevent possibility that
          the control program reads the In index between 8 bit system writes.
          Only need to do this if controller is in 8 bit mode.

          Optimizations are turned off when compiling this function to force
          the compiler to go to dual port memory each time the In index is read.
*/
WORD_T GetTxqIn(LINE_T far *pLn)
{
   WORD_T In;

   do
   {
      In = pLn->TxqIn;
   } while(In != pLn->TxqIn);
   return(In);
}

/***************************************************************************
Function: GetRxqOut
Purpose:  Get the Rx queue Out index.
Call:     GetRxqOut(pLn)
          LINE_T far *pLn; Ptr to line structure
Return:   WORD_T: The Rx queue Out index
Comments: This function repeatadly reads the Out index until the values from
          subsequent reads match.  This is done to prevent possibility that
          the control program reads the Out index between 8 bit system writes.
          Only need to do this if controller is in 8 bit mode.

          Optimizations are turned off when compiling this function to force
          the compiler to go to dual port memory each time the Out index is read.
*/
WORD_T GetRxqOut(LINE_T far *pLn)
{
   WORD_T Out;

   do
   {
      Out = pLn->RxqOut;
   } while(Out != pLn->RxqOut);
   return(Out);
}
#ifdef BORLANDC
#pragma option -O.
#endif
#ifdef MSC
#pragma optimize("",on)
#endif



