/******************************************************************************
 File:    dpram.h
 Purpose: Header file for sample C language programs for the HOSTESS i/S2.
          Contains dual port RAM address defines and data structures that
          must be known by both the system and the controller.
 Company: Comtrol Corporation
 Author:  Craig Harrison
 Ident:   $Id$
*******************************************************************************
 Copyright 1994 Comtrol Corporation. All rights reserved.
 Subject to developers license agreement.
******************************************************************************/

#ifndef DPRAM_H
#define DPRAM_H 1

/* ----- Fundemental Data Types -------------------------------------------- */
typedef char BYTE_T;
typedef unsigned char UBYTE_T;
typedef short WORD_T;
typedef unsigned short UWORD_T;
typedef long DWORD_T;
typedef unsigned long UDWORD_T;


/* ----- Dual port and I/O address and size definitions -------------------- */
/* Far memory addresses are given here in segment:offset format, for example:
   0xd0005000 = d000:5000 */
#define IO_SYS        0x218        /* SYS base I/O address */
#define ENTRY_SEG     0x1000       /* control program entry point segment */
#define ENTRY_OFF     0            /* control program entry point offset */
#define BASE_ADDR_SYS 0xd0000000L  /* SYS dual port RAM base address */
#define FIRMUSER_ADDR_SYS 0xd0000b80L /* SYS firmware user area addr */
#define LOAD_ADDR_SYS 0xd0000c00L /* SYS ctrlpgm download buffer address */
#define LOAD_ADDR_CTL 0x00000c00L /* CTL ctrlpgm download buffer address */
#define FIRMUSER_ADDR 0x00000b80L /* CTL firmware user area addr */

#define NUMLINES      2                /* number of serial lines */
#define VECTAREA_SIZE 0x400            /* vector area size */
#define FIRMDATA_SIZE 0x780            /* firmware data area size */
#define TXQ_SIZE      2048             /* Rx queue size */
#define RXQ_SIZE      2048             /* Rx queue size */
#define TXBQ_SIZE     8                /* number of bufs in Tx buffer queue */
#define RXBQ_SIZE     8                /* number of bufs in Rx buffer queue */
#define TXBQ_BUF_SIZE 1024             /* size of bufs in Tx buffer queue */
#define RXBQ_BUF_SIZE 1024+4           /* size of bufs in Rx buf q, +4 allows */
                                       /*   full 1K+CRC before DMA TC int */
#define SFG_SIZE      16               /* number of Status Flag Groups */
#define TRBUF_SIZE    0x800            /* trace buffer size */


/* ----- Message Queue data structure -------------------------------------- */
/* Used for controller and system messages. */
/* Message format:
   Byte 0 = Command.  These given by CTLCMD_X and SYSCMD_X #defines below.
   Byte 1 = Line number
   Byte 2-15 = Parameters as defined by command function.  These given by
               MSG_X_X #defines below.
*/
#define NUM_CTL_CMD  7                 /* number of SYS to CTL commands */
#define NUM_SYS_CMD  5                 /* number of CTL to SYS commands */
#define MSG_LEN   16                   /* number of bytes in one message */
#define MSGQ_SIZE 8                    /* number messages in queue buffer */
#define REPLY_LEN 7                    /* number bytes in reply */
typedef struct
{
   WORD_T In;                          /* add data here */
   WORD_T Out;                         /* remove data here */
   BYTE_T Buf[MSGQ_SIZE][MSG_LEN];     /* queue buffer */
   BYTE_T ReplyRdy;                    /* reply ready flag */
   BYTE_T ReplyBuf[REPLY_LEN];         /* reply buffer */
} MSGQ_T;

/* ----- Controller command numbers ---------------------------------------- */
/* This is always placed in byte 0 of the controller message, it identifies
   the function to execute. */
#define CTLCMD_TEST        0
#define CTLCMD_OPENASYNC   1
#define CTLCMD_OPENHDLC    2
#define CTLCMD_OPENBISYNC  3
#define CTLCMD_CLOSE       4
#define CTLCMD_GETMDM      5
#define CTLCMD_SETMDM      6

/* ----- System command numbers -------------------------------------------- */
/* This is always placed in byte 0 of the system message, it identifies
   the function to execute. */
#define SYSCMD_RXERROR        0
#define SYSCMD_TXERROR        1
#define SYSCMD_MDMCHG         2
#define SYSCMD_STATEERROR     4

/* ----- Flags used in controller & system messages ------------------------ */
/* General flags */
#define OFF        0
#define ON         1
#define NOCHANGE   2

/* Rx error flags */
#define MSG_RXERR_PARITY     0x01       /* parity error */
#define MSG_RXERR_FRAME      0x02       /* framing error */
#define MSG_RXERR_OVERRUN    0x04       /* overrun SCC Rx buffer */
#define MSG_RXERR_OVERFLOW   0x08       /* overflow dual port mem Rx buffer */
#define MSG_RXERR_ABORT      0x10       /* aborted Rx frame */
#define MSG_RXERR_CRC        0x20       /* bad CRC */
#define MSG_RXERR_BADCHAR    0x40       /* bad control char in BISYNC frame */

/* Tx error flags */
#define MSG_TXERR_RETRY      0x01       /* abort frame, retrying */
#define MSG_TXERR_RETRYMAX   0x02       /* attempts to send frame exceeded max
                                          retries */

/* Modem control flags */
#define MSG_MDM_RTS   0x01             /* request to send */
#define MSG_MDM_DTR   0x02             /* data terminal ready */
#define MSG_MDM_DSR   0x04             /* data set ready */
#define MSG_MDM_CD    0x08             /* carrier detect (CD in cpc.h) */
#define MSG_MDM_RI    0x10             /* ring indicator */
#define MSG_MDM_CTS   0x20             /* clear to send (CTS in cpc.h) */

/* Stop bit flags (used with async only) */
#define MSG_STOPBIT_1  0x0             /* 1 stop bit */
#define MSG_STOPBIT_15 0x1             /* 1.5 stop bits */
#define MSG_STOPBIT_2  0x2             /* 2 stop bits */

/* Data bit flags (used with async only) */
#define MSG_DATABIT_5  0x0             /* 5 data bits */
#define MSG_DATABIT_6  0x1             /* 6 data bits */
#define MSG_DATABIT_7  0x2             /* 7 data bits */
#define MSG_DATABIT_8  0x3             /* 8 data bits */

/* Parity flags (used with async only) */
#define MSG_PAR_NONE   0x0             /* no parity */
#define MSG_PAR_ODD    0x1             /* odd parity */
#define MSG_PAR_EVEN   0x2             /* even parity */

/* Baud rate flags (used with async only) */
#define MSG_BAUD_50     0x0
#define MSG_BAUD_75     0x1
#define MSG_BAUD_110    0x2
#define MSG_BAUD_134    0x3
#define MSG_BAUD_150    0x4
#define MSG_BAUD_300    0x5
#define MSG_BAUD_600    0x6
#define MSG_BAUD_1200   0x7
#define MSG_BAUD_1800   0x8
#define MSG_BAUD_2000   0x9
#define MSG_BAUD_2400   0xa
#define MSG_BAUD_3600   0xb
#define MSG_BAUD_4800   0xc
#define MSG_BAUD_7200   0xd
#define MSG_BAUD_9600   0xe
#define MSG_BAUD_19200  0xf
#define MSG_BAUD_38400  0x10
#define MSG_BAUD_57600  0x11
#define MSG_BAUD_76800  0x12
#define MSG_BAUD_115200 0x13

/* Clock mode flags (used with HDLC and Bisync) */
#define MSG_CLK_1 0x0     /* clock mode 1, both Tx & Rx clks DCE source */
#define MSG_CLK_2 0x1     /* clock mode 2, used for loopback testing. Tx clk
                             is DTE sourced from BRG, Rx clk DCE source. */


/* ----- Data buffer, used for queues of buffers --------------------------- */
#if defined(MSC)
#pragma pack(1)
#endif
typedef struct
{
   BYTE_T far *BufCtl;             /* ptr to buffer, CTL address */
   UWORD_T BufHi;                  /* hi word of buf abs addr for DMA ctrl */
   UWORD_T BufLo;                  /* lo word of buf abs addr for DMA ctrl */
   BYTE_T far *BufSys;             /* ptr to buffer, SYS address */
   UWORD_T Cnt;                    /* number data bytes in buffer */
} BUFFER_T;


/* ----- Line information structure ----------------------------------- */
/* There is one of these for each serial line. */
typedef struct
{
   WORD_T LineNum;                     /* line number */
   WORD_T BaseIo;                      /* SCC base I/O address */
   UWORD_T Status;                     /* line status (defined below) */
   WORD_T TxState;                     /* Tx state (defined below) */
   WORD_T RxState;                     /* Rx state (defined below) */
   WORD_T AltDtrMask;                  /* alternate DTR mask for set/clr DTR */
   WORD_T TxClkMask;                   /* mask to set Tx clk source DCE/DTE */
   UDWORD_T Timeout;                   /* CTL time at which timeout occurs */
   UBYTE_T Modem;                      /* state of modem control inputs */

   /* SCC register values */
   UBYTE_T WR2_;                       /* write register 2 */
   UBYTE_T WR3_;                       /* write register 3 */
   UBYTE_T WR4_;                       /* write register 4 */
   UBYTE_T WR5_;                       /* write register 5 */
   UBYTE_T WR12_;                      /* write register 12 */
   UBYTE_T WR13_;                      /* write register 13 */
   UBYTE_T WR15_;                      /* write register 15 */

   /* DMA channel info */
   BYTE_T TxDmaChan;                   /* DMA channel used for Tx */
   BYTE_T RxDmaChan;                   /* DMA channel used for Rx */
   BYTE_T EnTxDma;                     /* Tx DMA chan mask for enable */
   BYTE_T EnRxDma;                     /* Rx DMA chan mask for enable */
   BYTE_T DisTxDma;                    /* Tx DMA chan mask for disable */
   BYTE_T DisRxDma;                    /* Rx DMA chan mask for disable */

   /* Transmit queue info. This used for async data */
   WORD_T TxqIn;                       /* add data here */
   WORD_T TxqOut;                      /* remove data here */
   BYTE_T far *TxqCtl;                 /* ptr to Tx buffer, CTL CPU address */
   BYTE_T far *TxqSys;                 /* ptr to Tx buffer, SYS CPU address */

   /* Receive queue info. This used for async data */
   WORD_T RxqIn;                       /* add data here */
   WORD_T RxqOut;                      /* remove data here */
   BYTE_T far *RxqCtl;                 /* ptr to Tx buffer, CTL CPU address */
   BYTE_T far *RxqSys;                 /* ptr to Rx buffer, SYS CPU address */

   /* Transmit and receive buffer queue info (these are queues of buffers) */
   /* used for sync data */
   BYTE_T TxbqIn;                      /* add data here */
   BYTE_T TxbqOut;                     /* remove data here */
   BYTE_T TxbqCnt;                     /* number of buffers on the queue */
   BYTE_T RxbqIn;                      /* add data here */
   BYTE_T RxbqOut;                     /* remove data here */
   BYTE_T RxbqCnt;                     /* number of buffers on the queue */
   BYTE_T far *RxbqCurBuf;          /* ptr to current Rxbq buf (Bisync only) */
   WORD_T RxbqCurCnt;               /* current Rxbq buf count (Bisync only) */
   WORD_T TxbqRetryCnt;                /* number of retries on current buffer*/
   BUFFER_T Txb[TXBQ_SIZE];            /* buffer structures */
   BUFFER_T Rxb[RXBQ_SIZE];            /* buffer structures */

   UBYTE_T Fill2[10];                  /* filler */
} LINE_T;

/* ----- Line status definitions ------------------------------------------- */
/* These are used in LINE_T member Status */
#define ASYNC_OPEN  0x0001           /* line is open in async mode */
#define HDLC_OPEN   0x0002           /* line is open in HDLC mode */
#define BISYNC_OPEN 0x0004           /* line is open in BISYNC mode */
#define CLOSE_PEND  0x0008           /* a close on the line is pending */

/* ----- Synchronous Tx state definitions ---------------------------------- */
/* These are used in LINE_T member TxState.  They give the current state of
   the Tx state machine.  The same state names are used for both for HDLC and
   BISYNC. */
#define STX_IDLE       0              /* no Tx in process */
#define STX_ACTIVE     1              /* Tx DMA in process */
#define STX_RETRY      2              /* Tx frame aborted, retrying */
#define STX_CLOSE_WAIT 3         /* exceed max Tx retry, wait for cmdClose() */
#define STX_EOF_WAIT   4              /* Tx DMA done, waiting for EOF */
#define STX_EOF        5              /* Tx EOF processing */

/* ----- Rx state definitions ---------------------------------------------- */
/* These are used in LINE_T member RxState.  They give the current state of
   the Rx state machine.  There are seperate state names for HDLC and BISYNC,
   indicated by HRX_ and BRX_ prefixes respectively. */
#define HRX_BUF_WAIT   0              /* waiting for Rx buf available */
#define HRX_ACTIVE     1              /* Rx DMA in process */

#define BRX_BUF_WAIT   0              /* waiting for Rx buf available */
#define BRX_SOH_WAIT   1              /* waiting for SOH to start frame */
#define BRX_ACTIVE     2              /* receiving data in frame */
#define BRX_CRC1_WAIT  3              /* waiting first CRC byte */
#define BRX_CRC2_WAIT  4              /* waiting second CRC byte */
#define BRX_CRC3_WAIT  5              /* 1st post CRC byte time */
#define BRX_CRC4_WAIT  6              /* 2nd post CRC byte time */


/* ----- Firmware user area structure -------------------------------------- */
/* This is the same as defined by the firmware */
typedef struct
{
   UWORD_T i_flag;                     /* processor interaction flag */
   UWORD_T boot_flag;                  /* boot/activity flag */
   UWORD_T cfg_map;                    /* configuration map */
   BYTE_T fw_release[8];               /* firmware release number */
   BYTE_T sw_release[8];               /* control program release number */
   UDWORD_T dram_map1;                 /* DRAM map */
   UDWORD_T dram_map2;                 /* DRAM map */
   UDWORD_T scc_map;                   /* SCC map */
   UDWORD_T board_id;                  /* board ID */
   BYTE_T ii_flag;                     /* invalid interrupt flag */
   BYTE_T ii_type;                     /* invalid interrupt type */
   UWORD_T ii_cnt;                     /* invalid interrupt count */
   UDWORD_T  heartbeat;                /* heartbeat counter */
   BYTE_T cmd;                         /* firmware utility command */
   BYTE_T status;                      /* firmware utility status */
   BYTE_T msg[16];                     /* firmware utility message buffer */
   BYTE_T reserved1[0x40];             /* reserved for future use */
} FIRMUSER_T;

/* ----- Dual port semaphore ----------------------------------------------- */
typedef struct
{
   BYTE_T CtlAccess;
   BYTE_T SysAccess;
} DPSEM_T;

/* ----- Dual port memory structure ---------------------------------------- */
/* This pulls together all the components of lower 64K of dual port memory. */
typedef struct
{
   UBYTE_T VectorArea[VECTAREA_SIZE];  /* interrupt vectors */
   UBYTE_T FirmData[FIRMDATA_SIZE];    /* firmware data area - don't touch */
   FIRMUSER_T FirmUser;                /* firmware user area */
   BYTE_T TrBuf[TRBUF_SIZE];           /* trace buffer */
   WORD_T TrIx;                        /* trace buffer index */
   DPSEM_T DpSem;                      /* dual port semaphore */
   UWORD_T Sfg[SFG_SIZE];              /* SFGs */
   MSGQ_T CtlQ;                        /* CTL message queue */
   MSGQ_T SysQ;                        /* SYS message queue */
   BYTE_T Fill[4];                     /* fill for para alignment */
   LINE_T Line[NUMLINES];              /* line information */
   BYTE_T TxqBuf[NUMLINES][TXQ_SIZE];  /* Tx queue buffers */
   BYTE_T RxqBuf[NUMLINES][RXQ_SIZE];  /* Rx queue buffers */
   BYTE_T TxbqBuf[NUMLINES][TXBQ_SIZE][TXBQ_BUF_SIZE]; /* Tx buf q buffers */
   BYTE_T RxbqBuf[NUMLINES][RXBQ_SIZE][RXBQ_BUF_SIZE]; /* Rx buf q buffers */
   BYTE_T EndDPM;                      /* end of dual port memory usage */
} DPM_T;
#if defined(MSC)
#pragma pack()
#endif


#endif /* #ifndef DPRAM_H */
