/*****************************************************************************
File:    hilib.c
Purpose: C source for system routines to interface an application program
         to the Hostess i/S2 controller.
Company: Comtrol Corporation
Author:  Craig Harrison
Comment: This module is compiled with another C language source module to form
         a DOS application program that runs on the system CPU.  In the case
         of the tool kit programs diskette, HILIB.C and TEST.C are compiled
         to build TEST.EXE, a comprehensive test program used to demonstrate
         and check out the CPC.C control program.
Ident:   $Id$
*******************************************************************************
Copyright 1994 Comtrol Corporation. All rights reserved.
Subject to developers license agreement.
*******************************************************************************
                                 WARNINGS
1. The application program must call hiInitDpm() before calling any other
function in HILIB.C.

2. If using Microsoft(R) compilers make sure stack probes are disabled in any
function called by an ISR (Interrupt Service Routine).

3. When accessing a value from dual port memory using a non-atomic operation,
if that value can be modified by the control program that operation must be
protected. Some non-atomic read operations are:
   a. Double word reads
   b. Word reads when controller is in 8 bit mode (also called PC mode, see
      Control Register 1).
   c. Any read/write
A way to protect any non-atomic access is to use a semaphore (see hiGetDpSem()
and hiREL_DPSEM).A way to protect a non-atomic read is to repeatadly read the
value until subsequent reads match (see function hiReadAsync()).
******************************************************************************/

#include <dos.h>
#if defined(BORLANDC)
#include <mem.h>
#endif
#if defined(MSC)
#include <memory.h>
#endif
#include <conio.h>
#include <time.h>
#include "dpram.h"
#include "hilib.h"

MSGQ_T far *pCtlQ;                     /* ptr to controller message queue */
MSGQ_T far *pSysQ;                     /* ptr to system message queue */
UWORD_T far *pSfg;                     /* ptr to Status Flag Groups */
LINE_T far *Line[NUMLINES];            /* ptrs to line structures */
BYTE_T far *pDpSemCtlAccess;           /* ptr to dual port semaphore */
BYTE_T far *pDpSemSysAccess;           /* ptr to dual port semaphore */

IRQ_T far IrqInfo[] =                  /* IRQ information */
{
  {0xfe,0x8,0x21,0x00}, /* IRQ0 */
  {0xfd,0x9,0x21,0x00}, /* IRQ1 */
  {0xfb,0xa,0x21,0x00}, /* IRQ2 */
  {0xf7,0xb,0x21,0x08}, /* IRQ3 */
  {0xef,0xc,0x21,0x09}, /* IRQ4 */
  {0xdf,0xd,0x21,0x0a}, /* IRQ5 */
  {0xbf,0xe,0x21,0x00}, /* IRQ6 */
  {0x7f,0xf,0x21,0x00}, /* IRQ7 */
  {0xfe,0x70,0xa1,0x00}, /* IRQ8 */
  {0xfd,0x71,0xa1,0x0b}, /* IRQ9 */
  {0xfb,0x72,0xa1,0x0c}, /* IRQ10 */
  {0xf7,0x73,0xa1,0x0d}, /* IRQ11 */
  {0xef,0x74,0xa1,0x0e}, /* IRQ12 */
  {0xdf,0x75,0xa1,0x00}, /* IRQ13 */
  {0xbf,0x76,0xa1,0x00}, /* IRQ14 */
  {0x7f,0x77,0xa1,0x0f}  /* IRQ15 */
};
int IrqSave = 0;                         /* IRQ number save */
void (interrupt far *VectorSave)(void);  /* IRQ line interrupt vector save */
unsigned char MaskSave;                /* 8259 mask save */
WORD_T far EndOfIntIO;                 /* 8259 I/O address for EOI */


/* CR4 values for IRQ 0-15, 0 means IRQ is disabled */
char cr4[16] = {0,0,0,0x08,0x09,0x0a,0,0,0,0x0b,0x0c,0x0d,0x0e,0,0,0x0f};

/*****************************************************************************
Function: hiInitDpm
Purpose:  Initialize pointers to dual port memory structures.
Call:     hiInitDpm()
Return:   void
*/
void hiInitDpm(void)
{
   int Ln;                            /* line number */
   LINE_T far *pLn;                   /* ptr to line structure */
   DPM_T far *pDpm;                   /* ptr to dual port memory structure */
   int i;

   /* Initialize message queue pointers */
   pDpm = (DPM_T far *)BASE_ADDR_SYS;
   pCtlQ = &pDpm->CtlQ;
   pSysQ = &pDpm->SysQ;

   /* Initialize ptrs to dual port semaphore */
   pDpSemCtlAccess = &pDpm->DpSem.CtlAccess;
   pDpSemSysAccess = &pDpm->DpSem.SysAccess;

   /* Initialize ptr to Status Flag Groups */
   pSfg = pDpm->Sfg;

   /* Initialize line structures */
   for(Ln = 0;Ln < NUMLINES;Ln++)
   {
      pLn = &pDpm->Line[Ln];
      Line[Ln] = pLn;
      pLn->TxqSys = (BYTE_T far *)BASE_ADDR_SYS + (UWORD_T)pLn->TxqSys;
      pLn->RxqSys = (BYTE_T far *)BASE_ADDR_SYS + (UWORD_T)pLn->RxqSys;
      for(i = 0;i < TXBQ_SIZE;i++)
      {
         pLn->Txb[i].BufSys = (BYTE_T far *)BASE_ADDR_SYS +
                              (UWORD_T)pLn->Txb[i].BufSys;
      }
      for(i = 0;i < RXBQ_SIZE;i++)
      {
         pLn->Rxb[i].BufSys = (BYTE_T far *)BASE_ADDR_SYS +
                              (UWORD_T)pLn->Rxb[i].BufSys;
      }
   }
}

/*****************************************************************************
Function: hiEnqCtlMsg
Purpose:  Add a message to the controller message queue.
Call:     hiEnqCtlMsg(pMsg,pReply)
          BYTE_T *pMsg: Ptr to message to enqueue, must be MSG_LEN bytes long.
          BYTE_T *pReply: Ptr to buffer in which to place reply, if a reply
                        is expected.  Buffer must be at least REPLY_LEN bytes
                        long.  If no reply expected set to NULL.
Return:   int: 1 if message successfully enqueued and reply received (if
                 requested).
               0 if message could not be enqueued.
              -1 if time out occurred before a reply was received.
*/
int hiEnqCtlMsg(BYTE_T *pMsg,BYTE_T *pReply)
{
   WORD_T NumOpen;                    /* number open message spaces in queue */
   time_t t;                          /* system time at which timeout occurs */

   if((NumOpen = pCtlQ->Out - pCtlQ->In - 1) < 0) /* get open space in queue*/
      NumOpen += MSGQ_SIZE;            /* adjust for queue wrap */
   if(NumOpen <= 0)                    /* queue full */
      return(0);                       /* fail */

   _fmemcpy(pCtlQ->Buf[pCtlQ->In],pMsg,MSG_LEN);
   pCtlQ->In = (pCtlQ->In + 1) % MSGQ_SIZE; /* bump In pointer */
   OUTB(IO_SYS+2,0);                   /* interrupt controller */
   if(pReply != NULL)                  /* reply expected */
   {
      t = time(NULL) + REPLY_TIMEOUT;  /* time when for reply wait expires */
      while(time(NULL) <= t)
      {
         if(pCtlQ->ReplyRdy)           /* reply is ready */
         {
            _fmemcpy(pReply,pCtlQ->ReplyBuf,REPLY_LEN); /* copy reply buf */
            _fmemset(&pCtlQ->ReplyRdy,0,REPLY_LEN + 1); /* clr rdy flag & buf*/
            return(1);
         }
         else                          /* reply not ready */
         {
            hiDelay(4);                /* delay so don't slow down CTL CPU */
         }
      }
      return(-1);                      /* timeout before reply received */
   }
   else                                /* reply not expected */
   {
      return(1);                       /* success */
   }
}

/*****************************************************************************
Function: hiDeqSysMsg
Purpose:  Remove a message from the system message queue.
Call:     hiDeqSysMsg(pBuf)
          BYTE_T *pBuf: Ptr to buffer in which to place message.
Return:   int: 1 if message successfully dequeued.
               0 if no messages available.
*/
int hiDeqSysMsg(char *pBuf)
{
   if(pSysQ->Out == pSysQ->In)         /* no messages available */
      return(0);

   _fmemcpy(pBuf,pSysQ->Buf[pSysQ->Out],MSG_LEN); /* get message */
   pSysQ->Out = (pSysQ->Out + 1) % MSGQ_SIZE; /* bump out index */
   return(1);                          /* success */
}

/*****************************************************************************
Function: hiOpenAsync
Purpose:  Open a serial line for asynchronous communication.
Call:     hiOpenAsync(Ln,Baud,Parity,DataBit,StopBit,Modem)
          int Ln: Line number to open.
          int Baud: Baud rate flag, use only one of the MSG_BAUD_X flags.
          int Parity: Parity flag, use only one of the MSG_PAR_X flags.
          int DataBit: Number of data bits flag, use only one of the
                       MSG_DATABIT_X flags.
          int StopBit: Number of stop bits flag, use only one of the
                       MSG_STOPBIT_X flags.
          int Modem: Modem control output flags, use any combination of the
                     MSG_MDM_RTS and MSG_MDM_DTR flags.  If a flag is set
                     the corresponding modem control output will be made
                     active.
Return:   int: 1 if successful, 0 if failed
Comments: This function will complete successfully until the controller
          has completed the open.
*/
int hiOpenAsync(int Ln,int Baud,int Parity,int DataBit,int StopBit,int Modem)
{
   static BYTE_T Msg[MSG_LEN] =        /* controller message */
   {
      CTLCMD_OPENASYNC,                /* command number parameter */
      0,                               /* line number parameter */
      0,                               /* baud rate parameter */
      0,                               /* parity parameter */
      0,                               /* data bits parameter */
      0,                               /* stop bits parameter */
      0,                               /* modem control parameter */
      0,0,0,0,0,0,0,0,0      /* unused */
   };
   static BYTE_T ReplyBuf[REPLY_LEN];

   /* Verify parameters */
   if(Ln < 0 || Ln >= NUMLINES)
      return(0);
   if(Baud < MSG_BAUD_50 || Baud > MSG_BAUD_115200)
      return(0);
   if(Parity < MSG_PAR_NONE || Parity > MSG_PAR_EVEN)
      return(0);
   if(DataBit < MSG_DATABIT_5 || DataBit > MSG_DATABIT_8)
      return(0);
   if(StopBit < MSG_STOPBIT_1 || StopBit > MSG_STOPBIT_2)
      return(0);
   if(Modem < 0 || Modem > (MSG_MDM_DTR | MSG_MDM_RTS))
      return(0);

   /* Set up controller message */
   Msg[1] = (BYTE_T)Ln;                /* line number */
   Msg[2] = (BYTE_T)Baud;              /* baud rate flag */
   Msg[3] = (BYTE_T)Parity;            /* parity flag */
   Msg[4] = (BYTE_T)DataBit;           /* number data bits flag */
   Msg[5] = (BYTE_T)StopBit;           /* stop bits flag */
   Msg[6] = (BYTE_T)Modem;             /* modem control flags */

   /* Add message to controller message queue */
   if(hiEnqCtlMsg(Msg,ReplyBuf) == 1)
   {
      return(1);                  /* success */
   }
   else
      return(0);                  /* fail */
}

/*****************************************************************************
Function: hiOpenHdlc
Purpose:  Open a serial line for HDLC communication.
Call:     hiOpenHdlc(Ln,Modem,ClkMode)
          int Ln: Line number to open.
          int Modem: Modem control output flags, use any combination of the
                     MSG_MDM_RTS and MSG_MDM_DTR flags.  If a flag is set
                     the corresponding modem control output will be made
                     active.
          int ClkMode: Clock mode flag, use only one of the MSG_CLK_X flags.
          long DataRate: Data rate in BPS, only used if ClkMode = MSG_CLK_2.
Return:   int: 1 if successful, 0 if failed
Comments: This function will complete successfully until the controller
          has completed the open.
*/
int hiOpenHdlc(int Ln,int Modem,int ClkMode,long DataRate)
{
   static BYTE_T Msg[MSG_LEN] =        /* controller message */
   {
      CTLCMD_OPENHDLC,                 /* command number parameter */
      0,                               /* line number parameter */
      0,                               /* modem control parameter */
      0,                               /* clock mode parameter */
      0,0,0,0,                         /* data rate parameter */
      0,0,0,0,0,0,0,0                  /* unused */
   };
   DWORD_T *pDataRate;                 /* ptr to data rate field in Msg[] */
   static BYTE_T ReplyBuf[REPLY_LEN];

   /* Verify parameters */
   if(Ln < 0 || Ln >= NUMLINES)
      return(0);
   if(Modem < 0 || Modem > (MSG_MDM_DTR | MSG_MDM_RTS))
      return(0);
   if(ClkMode < MSG_CLK_1 || ClkMode > MSG_CLK_2)
      return(0);

   /* Set up controller message */
   Msg[1] = (BYTE_T)Ln;                /* line number */
   Msg[2] = (BYTE_T)Modem;             /* modem control flags */
   Msg[3] = (BYTE_T)ClkMode;           /* clock mode flag */
   pDataRate = (DWORD_T *)&Msg[4];     /* get ptr to 4 bytes for data rate */
   *pDataRate = DataRate;              /* data rate */

   /* Add message to controller message queue */
   if(hiEnqCtlMsg(Msg,ReplyBuf) == 1)
   {
      return(1);                  /* success */
   }
   else
      return(0);                  /* fail */
}

/*****************************************************************************
Function: hiOpenBisync
Purpose:  Open a serial line for Bisync communication.
Call:     hiOpenBisync(Ln,Modem,ClkMode)
          int Ln: Line number to open.
          int Modem: Modem control output flags, use any combination of the
                     MSG_MDM_RTS and MSG_MDM_DTR flags.  If a flag is set
                     the corresponding modem control output will be made
                     active.
          int ClkMode: Clock mode flag, use only one of the MSG_CLK_X flags.
          long DataRate: Data rate in BPS, only used if ClkMode = MSG_CLK_2.
Return:   int: 1 if successful, 0 if failed
Comments: This function will complete successfully until the controller
          has completed the open.
*/
int hiOpenBisync(int Ln,int Modem,int ClkMode,long DataRate)
{
   static BYTE_T Msg[MSG_LEN] =        /* controller message */
   {
      CTLCMD_OPENBISYNC,               /* command number parameter */
      0,                               /* line number parameter */
      0,                               /* modem control parameter */
      0,                               /* clock mode parameter */
      0,0,0,0,                         /* data rate parameter */
      0,0,0,0,0,0,0,0                  /* unused */
   };
   DWORD_T *pDataRate;                 /* ptr to data rate field in Msg[] */
   static BYTE_T ReplyBuf[REPLY_LEN];

   /* Verify parameters */
   if(Ln < 0 || Ln >= NUMLINES)
      return(0);
   if(Modem < 0 || Modem > (MSG_MDM_DTR | MSG_MDM_RTS))
      return(0);
   if(ClkMode < MSG_CLK_1 || ClkMode > MSG_CLK_2)
      return(0);

   /* Set up controller message */
   Msg[1] = (BYTE_T)Ln;                /* line number */
   Msg[2] = (BYTE_T)Modem;             /* modem control flags */
   Msg[3] = (BYTE_T)ClkMode;           /* clock mode flag */
   pDataRate = (DWORD_T *)&Msg[4];     /* get ptr to 4 bytes for data rate */
   *pDataRate = DataRate;              /* data rate */

   /* Add message to controller message queue */
   if(hiEnqCtlMsg(Msg,ReplyBuf) == 1)
   {
      return(1);                       /* success */
   }
   else
      return(0);                       /* fail */
}

/*****************************************************************************
Function: hiClose
Purpose:  Close a serial line.
Call:     hiClose(Ln)
          int Ln: Line number to close
Return:   int: 1 if successful, 0 if failed
*/
int hiClose(int Ln)
{
   static BYTE_T Msg[MSG_LEN] =        /* controller message */
   {
      CTLCMD_CLOSE,                    /* command number parameter */
      0,                               /* line number parameter */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0      /* unused */
   };

   Msg[1] = (BYTE_T)Ln;                /* set up line number in message */
   if(hiEnqCtlMsg(Msg,NULL) == 1)      /* add msg to controller msg que */
   {
      return(1);                       /* success */
   }
   else
      return(0);                       /* fail */
}

/*****************************************************************************
Function: hiGetModem
Purpose:  Get the modem control inputs
Call:     hiGetModem(Ln)
          int Ln: Line number
Return:   int: 0 if all modem control inputs are OFF.
               MSG_MDM_X flag for each modem control input that is ON, if any.
               -1 if failure.
*/
int hiGetModem(int Ln)
{
   static BYTE_T ReplyBuf[REPLY_LEN];
   static BYTE_T Msg[MSG_LEN] =
   {
      CTLCMD_GETMDM,                   /* command number parameter */
      0,                               /* line number parameter */
      0,0,0,0,0,0,0,0,0,0,0,0,0,0      /* unused */
   };

   if(Ln < 0 || Ln >= NUMLINES)
      return(-1);
   Msg[1] = (BYTE_T)Ln;                /* set up line number in message */
   if(hiEnqCtlMsg(Msg,ReplyBuf) == 1)  /* add message to controller msg que */
   {
      return(ReplyBuf[0]);             /* success */
   }
   else
      return(-1);                       /* fail */
}

/*****************************************************************************
Function: hiChangeModem
Purpose:  Change the modem control outputs
Call:     hiChangeModem(Ln)
          int Ln: Line number
          int DtrState: State to set DTR to, must be one of flags ON, OFF, or
                        NOCHANGE
          int RtsState: State to set RTS to, must be one of flags ON, OFF, or
                        NOCHANGE
Return:   int: 1 if successful, 0 if failed
*/
int hiChangeModem(int Ln,int DtrState,int RtsState)
{
   static BYTE_T Msg[MSG_LEN] =
   {
      CTLCMD_SETMDM,                   /* command number parameter */
      0,                               /* line number parameter */
      0,                               /* RTS state parameter */
      0,                               /* DTR state parameter */
      0,0,0,0,0,0,0,0,0,0,0,0      /* unused */
   };

   if(Ln < 0 || Ln >= NUMLINES)
      return(0);
   if(RtsState < OFF || RtsState > NOCHANGE)
      return(0);
   if(DtrState < OFF || DtrState > NOCHANGE)
      return(0);
   Msg[1] = (BYTE_T)Ln;                /* set up line number in message */
   Msg[2] = (BYTE_T)RtsState;          /* set up RTS state in message */
   Msg[3] = (BYTE_T)DtrState;          /* set up DTR state in message */
   if(hiEnqCtlMsg(Msg,NULL) == 1)      /* add message to controller msg que */
   {
      return(1);                       /* success */
   }
   else
      return(0);                       /* fail */
}

/*****************************************************************************
Function: hiMemCpy
Purpose:  Memory to memory copy for moving data into and out of dual port
          memory.
Call:     hiMemCpy(dest,src,n)
          void far *dest; Destination pointer
          void far *src; Source pointer
          size_t n; Number of bytes to copy
Return:   void
Comments: This function should be used to move blocks of data into and out
          of dual port memory rather than _fmemcpy().  This is because during
          a block move controller side bus accesses will operate at less than
          full bandwidth, which can delay processing of critical interrupts.
          To avoid hogging the controller side bus the number of bytes that
          this function moves in one operation is limited to DP_XFER_MAX, with
          a delay between copies. This is ugly, but necessary.
*/
void hiMemCpy(char far *dest,char far *src,unsigned n)
{
   unsigned i;

   while(n > 0)
   {
      if(n > DP_XFER_MAX)
         i = DP_XFER_MAX;
      else
         i = n;
      _fmemcpy(dest,src,i);
      n -= i;
      src += i;
      dest += i;
      hiDelay(2);
   }
}

/*****************************************************************************
Function: hiReadAsync
Purpose:  Read asynchronous data from a serial line.
Call:     hiReadAsync(Ln,pBuf,Cnt)
          int Ln: Line number to read from
          BYTE_T *pBuf: Ptr to buffer to place read data into
          int Cnt: Maximum number of bytes to read
Return:   int: Number of bytes read
*/
int hiReadAsync(int Ln,BYTE_T *pBuf,int Cnt)
{
   LINE_T far *pLn;                 /* ptr to line structure */
   int i;                           /* balance of chars to copy after q wrap */
   WORD_T Out, In;                  /* Rx queue In & Out indices */
   WORD_T cCnt;                     /* count of chars copied */

   pLn = Line[Ln];                  /* get ptr to line table entry */

   /* Get number of characters in Rx buffer */
   Out = pLn->RxqOut;               /* read Out */
   In = GetRxqIn(pLn);              /* read In */
   if((cCnt = In - Out) < 0)
      cCnt += RXQ_SIZE;                /* adjust for queue wrap */
   else if(cCnt == 0)
      return(cCnt);                    /* nothing in Rx buffer */
   if(Cnt < cCnt)
      cCnt = Cnt;                      /* don't overflow SYS buffer */
   i = cCnt - (RXQ_SIZE - Out);       /* i = whats left after wrap around */
   if(i < 0)
      i = 0;

   /* Copy to end of Rx buffer */
   hiMemCpy(pBuf,pLn->RxqSys + Out,cCnt - i);

   /* Point to beginning of buffer if already at end of it */
   Out = (Out + (cCnt - i)) % RXQ_SIZE;

   /* Copy the rest of the buffer, if any left */
   if (i != 0)
   {
      hiMemCpy(pBuf + (cCnt - i),pLn->RxqSys,i);
      Out = i;
   }

   /* Update Rx buffer Out */
   pLn->RxqOut = Out;
   return(cCnt);
}

/*****************************************************************************
Function: hiWriteAsync
Purpose:  Write asynchronous data to a serial line.
Call:     hiWriteAsync(Ln,pBuf,Cnt)
          int Ln: Line number to write to
          BYTE_T *pBuf: Ptr to buffer to get write data from
          int Cnt: Number of bytes to write
Return:   int: Number of bytes written
*/
int hiWriteAsync(int Ln,BYTE_T *pBuf,int Cnt)
{
   LINE_T far *pLn;                 /* ptr to line structure */
   int i;                           /* balance of chars to copy after q wrap */
   int NumOpen;                     /* num bytes open in Tx buffer */
   WORD_T In, Out;                  /* Tx queue In & Out indices */

   pLn = Line[Ln];                  /* get ptr to line table entry */
   In = pLn->TxqIn;                 /* read In */
   Out = GetTxqOut(pLn);            /* read Out */
   if((NumOpen = Out - In - 1) < 0) /* compute number open bytes */
      NumOpen += TXQ_SIZE;          /* adjust for q wrap */
   if(NumOpen > Cnt)
      NumOpen = Cnt;                /* don't move more than are incoming */
   if(NumOpen == 0)
      return(0);                    /* no room in Tx buffer */

   i = NumOpen - (TXQ_SIZE - In);   /* i = whats left after wrap around */
   if (i < 0)
      i = 0;

   /* Copy to end of Tx buffer */
   hiMemCpy(pLn->TxqSys + In,pBuf,NumOpen - i);

   /* Point to beginning of buffer if already at end of it */
   In = (In + (NumOpen - i)) % TXQ_SIZE;

   /* Copy the rest of the buffer, if any left */
   if (i != 0)
   {
      hiMemCpy(pLn->TxqSys,pBuf + (NumOpen - i),i);
      In = i;
   }

   /* Update Tx buffer In */
   pLn->TxqIn = In;
   return(NumOpen);
}

/*****************************************************************************
Function: hiGetRxByteCnt
Purpose:  Get the number of bytes currently queued on the Rx queue.
Call:     hiGetRxByteCnt(Ln)
          int Ln: Line number
Return:   int: The number of Rx data bytes currently queued.
Comments: The Rx queue is used for asynchronous data.  Do not confuse this
          with the Rx buffer queue, which is used for synchronous data.
*/
int hiGetRxByteCnt(int Ln)
{
   LINE_T far *pLn;                    /* ptr to line structure */
   WORD_T Out, In;                     /* Rx queue In & Out indices */
   WORD_T Cnt;                         /* queue byte count */

   pLn = Line[Ln];                     /* get ptr to line table entry */
   Out = pLn->RxqOut;                  /* read Out */
   In = GetRxqIn(pLn);                 /* read In */
   if((Cnt = In - Out) < 0)            /* compute number bytes */
      Cnt += RXQ_SIZE;                 /* adjust for queue wrap */
   return(Cnt);
}

/*****************************************************************************
Function: hiGetTxByteCnt
Purpose:  Get the number of bytes currently queued on the Tx queue.
Call:     hiGetTxByteCnt(Ln)
          int Ln: Line number
Return:   int: The number of Tx data bytes currently queued.
Comments: The Tx queue is used for asynchronous data.  Do not confuse this
          with the Tx buffer queue, which is used for synchronous data.
*/
int hiGetTxByteCnt(int Ln)
{
   LINE_T far *pLn;                    /* ptr to line structure */
   WORD_T Out, In;                     /* Tx buffer In & Out indices */
   WORD_T Cnt;                         /* queue byte count */

   pLn = Line[Ln];                     /* get ptr to line table entry */
   In = pLn->TxqIn;                    /* read In */
   Out = GetTxqOut(pLn);               /* read Out */
   if((Cnt = In - Out) < 0)            /* compute number bytes */
      Cnt += RXQ_SIZE;                 /* adjust for queue wrap */
   return(Cnt);
}

/*****************************************************************************
Function: hiWriteSync
Purpose:  Write synchronous data to a serial line.  This is used for HDLC
          and Bisync, it writes an entire frame of data.
Call:     hiWriteSync(Ln,pBuf,Cnt)
          int Ln: Line number to write to
          BYTE_T *pBuf: Ptr to buffer to get write data from
          int Cnt: Number of bytes to write, cannot exceed TXBQ_BUF_SIZE.
Return:   int: 1 if successful, 0 if failed because no controller Tx buffer
               available.
Comments: If a controller Tx buffer is available the entire contents of pBuf
          is moved into the buffer, up to TXBQ_BUF_SIZE.
*/
int hiWriteSync(int Ln,BYTE_T *pBuf,int Cnt)
{
   LINE_T far *pLn;                 /* ptr to line structure */
   BYTE_T In;                       /* Tx buffer queue In index */

   pLn = Line[Ln];                  /* get ptr to line table entry */
   if(pLn->TxbqCnt >= TXBQ_SIZE)    /* no buffer available */
      return(0);
   if(Cnt > TXBQ_BUF_SIZE)          /* don't exceed size of buffer */
      Cnt = TXBQ_BUF_SIZE;
   In = pLn->TxbqIn;                /* read In */

   pLn->Txb[In].Cnt = Cnt;             /* update count in buffer struct */
   hiMemCpy(pLn->Txb[In].BufSys,pBuf,Cnt); /* copy data to DP memory */

   /* Update Tx buffer queue In index and count.  Need to protect TxbqCnt++
      with dual port semaphore because the ++ operation is not atomic and
      there is the possibliity that both system and controller could be
      changing it simultaneously. */
   _disable();
   hiGetDpSem();
   pLn->TxbqIn = (In + 1) % TXBQ_SIZE;
   pLn->TxbqCnt++;
   hiREL_DPSEM();
   _enable();
   return(1);
}

/*****************************************************************************
Function: hiReadSync
Purpose:  Read synchronous data from a serial line.  This is used for HDLC
          and Bisync, it reads an entire frame of data.
Call:     hiReadSync(Ln,pBuf)
          int Ln: Line number to read from
          BYTE_T *pBuf: Ptr to buffer to put read data into.  This buffer
             must be at least RXBQ_BUF_SIZE bytes in size.
Return:   int:  0: No data available to read.
               >0: Read successful, is the number of data bytes in the
                   frame read into pBuf.
*/
int hiReadSync(int Ln,BYTE_T *pBuf)
{
   LINE_T far *pLn;                 /* ptr to line structure */
   BYTE_T Out;                      /* Rx buffer Out index */
   WORD_T Cnt;                      /* number of bytes in controller buffer */

   pLn = Line[Ln];                  /* get ptr to line table entry */
   if(pLn->RxbqCnt <= 0)            /* no buffer available */
   {
      return(0);
   }
   Out = pLn->RxbqOut;              /* read Out */

   Cnt = pLn->Rxb[Out].Cnt;         /* get number of bytes in buffer struct */
   hiMemCpy(pBuf,pLn->Rxb[Out].BufSys,Cnt); /* copy data from DP memory */

   /* Update Rx buffer queue Out index and count.  Need to protect RxbqCnt--
      with dual port semaphore because the ++ operation is not atomic and
      there is the possibliity that both system and controller could be
      changing it simultaneously. */
   pLn->RxbqOut = (Out + 1) % RXBQ_SIZE;
   _disable();
   hiGetDpSem();
   pLn->RxbqCnt--;
   hiREL_DPSEM();
   _enable();

   return(Cnt);
}

/******************************************************************************
Function: hiInitSfg
Purpose:  Initialize all flags in a Status Flag Group.
Call:     hiInitSfg(Group,Value)
          int Group: Status flag group number, identifies which SFG
                     to operate on.
          UWORD_T Value: The value to initialize the SFG to.
Return:   void
Comments: An SFG is a word location in dual port memory which may be read or
          written by either the system or the controller.  Each bit within the
          SFG can be thought of as a seperate flag, so there are 16 flags
          in each SFG.  The number of SFGs that exist is defined by SFG_SIZE.
          SFGs are identified by group number.  Group numbers range from 0 to
          SFG_SIZE - 1.  SFGs are general purpose flag groups.  They can be
          used for intertask or interprocessor communication of any status
          desired.  All functions that operate on SFGs block other tasks
          and the obtain exclusive dual port memory access before accessing the
          SFG until the operation is complete.  This protects the caller
          against races in case multiple tasks or CPUs try to access an SFG
          simultaneusly. The following functions are available for
          manipulating SFGs:  hiClrSfg(), hiDecrementSfg(), hiGetSfg(),
          hiIncrementSfg(), hiInitSfg(), hiSetSfg(), and hiTestSetSfg().  The
          controller has analagous functions for operating on SFGs.
Warnings: This function disables and enables interrupts.
*/
void hiInitSfg(int Group,UWORD_T Value)
{
  _disable();
  hiGetDpSem();
  pSfg[Group] = Value;
  hiREL_DPSEM();
  _enable();
}

/******************************************************************************
Function: hiGetSfg
Purpose:  Get a masked copy of a Status Flag Group.
Call:     sGetMaskedSFG(Group,Mask)
          int Group: Status flag group number, identifies which SFG
                     to operate on.
          UWORD_T Mask:  Only the flags set in Mask have their states
                     returned, all flags not set in Mask are cleared
                     in the return value.   If the current state of all
                     flags is desired set Mask to ffffH.
Return:   UWORD_T: The masked copy of the SFG.
Comments: See hiInitSfg().
Warnings: This function disables and enables interrupts.
*/
UWORD_T hiGetSfg(int Group,UWORD_T Mask)
{
   UWORD_T State;                      /* the state of the SFG */

   _disable();
   hiGetDpSem();
   State = pSfg[Group];
   hiREL_DPSEM();
   _enable();
   return(State & Mask);
}

/******************************************************************************
Function: hiTestSetSfg
Purpose:  Test and set the state of a Status Flag Group.
Call:     hiTestSetSfg(Group,Mask)
          int Group: Status flag group number, identifies which SFG
                     to operate on.
          UWORD_T Mask:  Or mask.  All flags set to 1 in Mask will be set in
                         the flag group, other flags will be unchanged.
Return:   UWORD_T: The state of the SFG before the set operation was
                   performed.
Comments: This function first saves the state of the SFG, then performs the
          set operation, then returns the saved state.
          See hiInitSfg().
Warnings: This function disables and enables interrupts.

*/
UWORD_T hiTestSetSfg(int Group,UWORD_T Mask)
{
   UWORD_T State;                      /* state of flag group */

   _disable();
   hiGetDpSem();
   State = pSfg[Group];
   pSfg[Group] |= Mask;
   hiREL_DPSEM();
   _enable();
   return(State);
}

/******************************************************************************
Function: hiIncrementSfg
Purpose:  Increment an SFG and return it's state after incrementing.
Call:     hiIncrementSfg(Group)
          int Group: Status flag group number, identifies which SFG to operate
                     on.
Return:   UWORD_T: The state of the SFG after incrementing.
Comments: This function is useful if an SFG is to be used to count things.
          In this case the SFG would be thought of as a UWORD_T value rather
          than a group of 16 flags.
          See hiInitSfg().
Warnings: This function disables and enables interrupts.
*/
UWORD_T hiIncrementSfg(int Group)
{
   UWORD_T State;                      /* the state of the SFG */

   _disable();
   hiGetDpSem();
   State = ++pSfg[Group];
   hiREL_DPSEM();
   _enable();
   return(State);
}

/******************************************************************************
Function: hiDecrementSfg
Purpose:  Decrement an SFG and return it's state after decrementing.
Call:     hiDecrementSfg(Group)
          int Group: Status flag group number, identifies which SFG to operate
                     on.
Return:   UWORD_T: The state of the SFG after decrementing.
Comments: This function is useful if an SFG is to be used to count things.
          In this case the SFG would be thought of as a UWORD_T value rather
          than a group of 16 flags.
          See hiInitSfg().
Warnings: This function disables and enables interrupts.
*/
UWORD_T hiDecrementSfg(int Group)
{
   UWORD_T State;                      /* the state of the SFG */

   _disable();
   hiGetDpSem();
   State = --pSfg[Group];
   hiREL_DPSEM();
   _enable();
   return(State);
}

/******************************************************************************
Function: hiClrSfg
Purpose:  Clear one or more flags in a Status Flag Group.
Call:     hiClrSfg(Group,Mask)
          int Group: Status flag group number, identifies which SFG to operate
                     on.
          UWORD_T Mask: Clear mask.  All flags set to 1 in Mask will be
                        cleared, other flags will be unchanged.
Return:   void
Comments: See hiInitSfg().
Warnings: This function disables and enables interrupts.
*/
void hiClrSfg(int Group,UWORD_T Mask)
{
   _disable();
   hiGetDpSem();
   pSfg[Group] &= ~Mask;
   hiREL_DPSEM();
   _enable();
}

/******************************************************************************
Function: hiSetSfg
Purpose:  Set one or more flags in a Status Flag Group.
Call:     hiSetSfg(Group,Mask)
          int Group: Status flag group number, identifies which SFG to operate
                     on.
          UWORD_T Mask: Or mask.  All flags set to 1 in Mask will be set in the
                        flag group, other flags will be unchanged.
Return:   void
Comments: See hiInitSfg().
Warnings: This function disables and enables interrupts.
*/
void hiSetSfg(int Group,UWORD_T Mask)
{
  _disable();
  hiGetDpSem();
  pSfg[Group] |= Mask;
  hiREL_DPSEM();
  _enable();

}

/*****************************************************************************
Function:   hiGetDpSem
Purpose:    Get dual port semaphore.
Comments:   This function is used to guarentee the system will have exclusive
            access to a portion of dual port memory until hiREL_DPSEM()
            is called.  Normally this is only used to protect critical portions
            of memory that can be operated on by both the controller and the
            system.  For this to work properly the controller must also also get
            and release the same semaphore using the companion macros
            GET_DPSEM and REL_DPSEM.
Warnings:   Interrupts must be disabled before calling this function.
            hiGetDpSem and hiREL_DPSEM must always be used in pairs.
            The time that the semaphore is held must be kept as short
            as possible because the controller CPU will spin while it is
            waiting for the semaphore.
*/
void hiGetDpSem(void)
{
   while(1)                            /* do until access granted */
   {
      if(!*pDpSemCtlAccess)            /* controller does not have access */
      {
         *pDpSemSysAccess = 1;         /* indicate system wants access */
         if(!*pDpSemCtlAccess)         /* system access granted */
            break;
         else                          /* controller got access in the meantime */
            *pDpSemSysAccess = 0;      /* give it to controller */
      }
      hiDelay(2);                      /* delay so don't slow down CTL CPU */
   }
}

/*****************************************************************************
Function: hiDelay
Purpose:  Waste some time.
Call:     hiDelay(int Cycles)
          long Cycles: Number of I/O bus cycles to wait.  Each cycle takes
                       about 300 nanoseconds on an ISA machine.
Return:   void
Comments: If system and controller are both attempting to accessing dual port
          memory the two CPUs will alternate in gaining access.  Since the
          controller CPU must use the dual port memory bus for all it's
          memory accesses (both code and data), heavy system dual port memory
          usage can slow the controller CPU to half speed.  This can
          delay controller processing of critical interrupts.  This function
          can be called between system memory accesses to allow the controller
          CPU to run faster.
*/
void hiDelay(long Cycles)
{
   long i;

   for(i = 0;i < Cycles;i++)
      INB(IO_SYS);                     /* dummy I/O read to generate delay */
}

/*******************************************************************************
Function: hiInstallIrq
Purpose:  Install an Interrupt Request vector and enable the interrupt.
Call:     hiInstallIrq(Irq,IntHnd)
          int Irq; Interrupt request line number, must be an IRQ
                   supported by the controller.
          void (interrupt far *IntHnd)(); Ptr to interrupt handler
Return:   void
Comments: This functions saves the old interrupt vector, installs the new
          vector, saves the previous state of the 8259 mask register bit for
          this IRQ, enables this IRQ in the 8259 mask register, and enables
          the IRQ on the controller.

          If the IRQ is not supported by the controller it will not be
          installed.
Warnings: After calling hiInstallIrq(), hiRestoreIrq() must be called before
          calling hiInstallIrq() again.

          If the application calls hiInstallIrq(), hiRestoreIrq() should
          be called before exiting the application.  Failure to do so could
          result in a system crash because an interrupt to the system could
          occur, but the interrupt handler would no longer exist.
*/
void hiInstallIrq(int Irq,void (interrupt far *IntHnd)())
{
   IRQ_T far *pIrq;                    /* ptr to IRQ information structure */
   unsigned char Mask;                 /* 8259 mask */

   if(Irq < 0 || Irq > 15)             /* IRQ out of range */
      return;
   pIrq = &IrqInfo[Irq];               /* get ptr to IRQ info structure */
   if(pIrq->Cr4 == 0)                  /* IRQ not supported by this CTL */
      return;
   IrqSave = Irq;                      /* save this IRQ number */
   EndOfIntIO = pIrq->MaskIO - 1;      /* save EOI address */

   /* Save vector.  The cast is needed to avaoid a C 7.0 warning that
      I can't get rid of otherwise. */
   VectorSave = (void (interrupt far *)(void))_dos_getvect(pIrq->IntNum);

   hiDIS_INT();                        /* disable interrupts */
   _dos_setvect(pIrq->IntNum,IntHnd);  /* put new vector in vector table */
   Mask = INB(pIrq->MaskIO);           /* get current 8259 mask register */
   MaskSave = Mask & ~pIrq->Mask;      /* save bit for IRQ line, others low */
   Mask &= pIrq->Mask;                 /* enable IRQ line */
   OUTB(pIrq->MaskIO,Mask);            /* write it to 8259 */
   outp(IO_SYS+1,CR4 | EN_RAM); /* access CTL CR4, keep dual port mem enabled*/
   outp(IO_SYS,pIrq->Cr4);             /* enable IRQ on controller */
   hiEN_INT();                         /* enable interrupts */
}

/*******************************************************************************
Function: hiRestoreIrq
Purpose:  Restore an IRQ previously installed with hiInstallIrq() to it's
          original state.
Call:     hiRestoreIrq(Irq)
          int Irq; Interrupt request line number
Return:   void
Comments: This functions restores the old interrupt vector, restores
          the previous 8259 mask register bit for this IRQ line, and
          disables the IRQ on the controller.
Warnings: If the application has called hiInstallIrq(), this function should
          be called before exiting the application.  Failure to do so could
          result in a system crash because an interrupt to the system could
          occur, but the interrupt handler would no longer exist.
*/
void hiRestoreIrq()
{
   IRQ_T far *pIrq;                    /* ptr to IRQ information structure */
   unsigned char Mask;                 /* 8259 mask */

   if(IrqSave == 0)                    /* if vector never installed */
      return;                          /* don't reinstall it */
   pIrq = &IrqInfo[IrqSave];           /* get ptr to IRQ info structure */
   hiDIS_INT();                        /* disable interrupts */
   outp(IO_SYS+1,CR4 | EN_RAM); /* access CTL CR4, keep dual port mem enabled*/
   outp(IO_SYS,0);                     /* disable IRQ on controller */
   _dos_setvect(pIrq->IntNum,VectorSave);

   Mask = INB(pIrq->MaskIO);           /* get current 8259 mask register */
   Mask |= MaskSave;                   /* restore bit for IRQ, others same */
   OUTB(pIrq->MaskIO,Mask);            /* write it to 8259 */
   IrqSave = 0;                        /* indicate our vector not installed */
   hiEN_INT();                         /* enable interrupts */
}

#ifdef MSC
#pragma check_stack( off ) /* these used in ISR so stack probes disabled */
#endif
/*****************************************************************************
Function: hiGetTxBufCnt
Purpose:  Get the number of Tx buffers currently queued on the Tx buffer queue.
Call:     hiGetTxBufCnt(Ln)
          int Ln: Line number
Return:   int: The number of buffers queued.
Comments: The Tx buffer queue is used for synchronous data.  Do not confuse
          this with the Tx queue, which is used for asynchronous data.
*/
int hiGetTxBufCnt(int Ln)
{
   return(Line[Ln]->TxbqCnt);
}

/*****************************************************************************
Function: hiGetRxBufCnt
Purpose:  Get the number of Rx buffers currently queued on the Rx buffer queue.
Call:     hiGetRxBufCnt(Ln)
          int Ln: Line number
Return:   int: The number of buffers queued.
Comments: The Rx buffer queue is used for synchronous data.  Do not confuse
          this with the Rx queue, which is used for asynchronous data.
*/
int hiGetRxBufCnt(int Ln)
{
   return(Line[Ln]->RxbqCnt);
}

/*****************************************************************************
Function: hiGetSysMsgStatus
Purpose:  Get status of system message queue
Call:     hiGetSysMsgStatus(pBuf)
          BYTE_T *pBuf: Ptr to buffer in which to place message.
Return:   int: 1 if message(s) available
               0 if no messages available.
*/
int hiGetSysMsgStatus(void)
{
   if(pSysQ->Out == pSysQ->In)         /* no messages available */
      return(0);
   else
      return(1);
}

/*******************************************************************************
Function: hiEOI
Purpose:  End of interrupt command to PIC (Peripheral Interrupt Controller)
Call:     EndOfInt()
Return:   void
Comments: This must be called before exiting an interrupt handler.
*/
void hiEndOfInterrupt(void)
{
   OUTB(EndOfIntIO,0x20);
   if(EndOfIntIO == 0xa0)
      OUTB(0x20,0x20);
}


#ifdef BORLANDC
#pragma option -Od
#endif
#ifdef MSC
#pragma optimize("",off)
#endif
/***************************************************************************
Function: GetTxqOut
Purpose:  Get the Tx queue Out index.
Call:     GetTxqOut(pLn)
          LINE_T far *pLn; Ptr to line structure
Return:   WORD_T: The Tx queue Out index
Comments: This function repeatadly reads the Out index until the values from
          subsequent reads match.  This is done to prevent possibility that
          the control program reads the Out index between 8 bit system writes.
          Only need to do this if controller is in 8 bit mode.

          Optimizations are turned off when compiling this function to force
          the compiler to go to dual port memory each time the Out index is read.
*/
WORD_T GetTxqOut(LINE_T far *pLn)
{
   WORD_T Out;

   do
   {
      Out = pLn->TxqOut;
   } while(Out != pLn->TxqOut);
   return(Out);
}

/***************************************************************************
Function: GetRxqIn
Purpose:  Get the Rx queue In index.
Call:     GetRxqIn(pLn)
          LINE_T far *pLn; Ptr to line structure
Return:   WORD_T: The Rx queue In index
Comments: This function repeatadly reads the In index until the values from
          subsequent reads match.  This is done to prevent possibility that
          the control program reads the In index between 8 bit system writes.
          Only need to do this if controller is in 8 bit mode.

          Optimizations are turned off when compiling this function to force
          the compiler to go to dual port memory each time the In index is read.
*/
WORD_T GetRxqIn(LINE_T far *pLn)
{
   WORD_T In;

   do
   {
      In = pLn->RxqIn;
   } while(In != pLn->RxqIn);
   return(In);
}
#ifdef BORLANDC
#pragma option -O.
#endif
#ifdef MSC
#pragma optimize("",on)
#pragma check_stack( on )
#endif











